/*****************************************************************/
/*  ELEC 999                                                     */
/*  University of Nebraska-Lincoln                               */
/*  Department of Electrical Engineering                         */
/*  David J. Russell                                             */
/*  1-29-09                                                      */
/*                                                               */
/* 01/30/09 - Added % distribution to output type, version 1.1   */
/* 01/31/09 - Added windowing across an input, version 1.2       */
/* 02/03/09 - Added unpaired repeats and modified paired counts, */
/*             version 1.3                                       */
/* 02/14/09 - Added html output type where text is colored to    */
/*             identify the structural pieces. Modified          */
/*             VectorADT to allow longer vectors, version 1.5    */
/*****************************************************************/

#define _DIST

/*********************************************************************************/
/* Included Header Files */
/*********************************************************************************/
#include "Globals.h"
#include "VectorADT.h"
#include "CNFGrammarADT.h"
#include "Distribution.h"
#include "Parameters.h"

/*********************************************************************************/
/* Private Module Constants */
/*********************************************************************************/
#define DIST_HEAD_ARRAY_CHUNK_SIZE  10

/*********************************************************************************/
/* Private Type Definitions */
/*********************************************************************************/
typedef struct
{
   ULONG InvertedDepth;
   ULONG UnpairedRepeatCount;
   ULONG UnpairedComplementCount;
   ULONG PairedRepeatCount;
   ULONG LastHeadRepeatIndex;
   ULONG LastComplementIndex;
} DIST_HEAD_TYPE;

typedef struct
{
   ULONG Length;
   DIST_HEAD_TYPE *Cells;
   ULONG SequencePosition;
} DIST_HEAD_ARRAY_TYPE;

typedef struct
{
   ULONG Length;
   ULONG *PieceHead;
   ULONG SequencePosition;
} DIST_PIECES_ARRAY_TYPE;

/*********************************************************************************/
/* Private Module Function Prototypes */
/*********************************************************************************/
static void DIST_InitializeHeadArray(void);
static void DIST_CleanUpHeadArray(void);
static void DIST_InitializePiecesArray(ULONG Length);
static void DIST_SetHeadArrayElement(CNFADT_HEAD_TYPE Head, 
                                       ULONG InvertedDepth, 
                                       BOOLEAN ReverseComplement);
static void DIST_SetPiecesArrayElement(CNFADT_HEAD_TYPE Head, 
                                       ULONG InvertedDepth);
static void DIST_ScanForStructuralPiece(CNFADT_DICTIONARY_TYPE *Dictionary, 
                                          CNFADT_HEAD_TYPE Head,
                                          BOOLEAN ReverseComplement);
static void DIST_ScanHeadArrayForStructuralCoverage(ULONG *TotalUnpairedRepeats,
                                                    ULONG *TotalPaired);
static void DIST_ScanPiecesArrayForStructuralMultiplicity(void);

/*********************************************************************************/
/* Private Module Variables */
/*********************************************************************************/
static DIST_HEAD_ARRAY_TYPE m_HeadArray;
static DIST_PIECES_ARRAY_TYPE m_PiecesArray;

/*********************************************************************************/
/* Public and Private Module Functions */
/*********************************************************************************/

void DIST_GenerateStructureDistribution(CNFADT_DICTIONARY_TYPE *Dictionary, 
                                          CNFADT_HEAD_TYPE Head,
                                          ULONG SequenceLength,
                                          FLOAT *UnpairedRepeatPercentage,
                                          FLOAT *PairedPercentage)
{
   ULONG TotalUnpairedRepeats;
   ULONG TotalPaired;

   DIST_InitializeHeadArray();
   DIST_InitializePiecesArray(SequenceLength);
   
   DIST_ScanForStructuralPiece(Dictionary, Head, FALSE);
   DIST_ScanHeadArrayForStructuralCoverage(&TotalUnpairedRepeats, 
                                             &TotalPaired);
   DIST_ScanPiecesArrayForStructuralMultiplicity();

   DIST_CleanUpHeadArray();  
   /* 
      WARNING:
      
      We don't clean up the pieces array here so that the
      HTML write stuff can use the information.
      
      So, we need to call it externally after this funciton is called,
      and the data is no longer needed.
   */

   *UnpairedRepeatPercentage = (((FLOAT) TotalUnpairedRepeats) / ((FLOAT) SequenceLength)) * 100;
   *PairedPercentage = (((FLOAT) TotalPaired) / ((FLOAT) SequenceLength)) * 100;
}

/*********************************************************************************/

void DIST_InitializeHeadArray(void)
{
   m_HeadArray.Length = 0;
   m_HeadArray.Cells = NULL;
   m_HeadArray.SequencePosition = 0;
}

/*********************************************************************************/

void DIST_CleanUpHeadArray(void)
{
   if (m_HeadArray.Cells != NULL)
   {
      free(m_HeadArray.Cells);
      m_HeadArray.Cells = NULL;
   }

   m_HeadArray.Length = 0;
}

/*********************************************************************************/

void DIST_InitializePiecesArray(ULONG Length)
{
   m_PiecesArray.PieceHead = (ULONG *) malloc(sizeof(ULONG) * Length);
   if (m_PiecesArray.PieceHead == NULL)
   {
      printf("Error: malloc failed in DIST_InitializePiecesArray.\n");
      exit(1);
   }
   m_PiecesArray.Length = Length;
   m_PiecesArray.SequencePosition = 0;
   memset (m_PiecesArray.PieceHead, 0, sizeof(ULONG) * Length);
}

/*********************************************************************************/

void DIST_CleanUpPiecesArray(void)
{
   if (m_PiecesArray.PieceHead != NULL)
   {
      free(m_PiecesArray.PieceHead);
      m_PiecesArray.PieceHead = NULL;
   }
   m_PiecesArray.Length = 0;
}

/*********************************************************************************/

ULONG DIST_GetPiecesArray(ULONG **Array)
{
   *Array = m_PiecesArray.PieceHead;
   return (m_PiecesArray.Length);
}

/*********************************************************************************/

void DIST_SetHeadArrayElement(CNFADT_HEAD_TYPE Head, 
                              ULONG InvertedDepth, 
                              BOOLEAN ReverseComplement)
{
   BOOLEAN CheckNextCondition;
   
   while (Head >= m_HeadArray.Length)
   {
      if (m_HeadArray.Length == 0)
      {
         m_HeadArray.Cells = (DIST_HEAD_TYPE *) malloc(sizeof(DIST_HEAD_TYPE) * DIST_HEAD_ARRAY_CHUNK_SIZE);

         if (m_HeadArray.Cells == NULL)
         {
            printf("Error: malloc failed in DIST_SetHeadArrayElement.\n");
            exit(1);
         }
      }
      else
      {
         m_HeadArray.Cells = (DIST_HEAD_TYPE *) realloc(m_HeadArray.Cells, (m_HeadArray.Length + DIST_HEAD_ARRAY_CHUNK_SIZE) * sizeof(DIST_HEAD_TYPE));

         if (m_HeadArray.Cells == NULL)
         {
            printf("Error: realloc failed in DIST_SetHeadArrayElement.\n");
            exit(1);
         }
      }

      memset(&(m_HeadArray.Cells[m_HeadArray.Length]), 0, DIST_HEAD_ARRAY_CHUNK_SIZE * sizeof(DIST_HEAD_TYPE));
      m_HeadArray.Length += DIST_HEAD_ARRAY_CHUNK_SIZE;
   }

   m_HeadArray.Cells[Head].InvertedDepth = InvertedDepth;
   if (ReverseComplement == TRUE)
   {
      CheckNextCondition = TRUE;
      if (m_HeadArray.Cells[Head].UnpairedRepeatCount > 0)
      {
         if ((m_HeadArray.SequencePosition - m_HeadArray.Cells[Head].LastHeadRepeatIndex) < PARMS_GetPairedDistanceThreshold())
         {
            m_HeadArray.Cells[Head].PairedRepeatCount += 1;
            m_HeadArray.Cells[Head].UnpairedRepeatCount -= 1;
            CheckNextCondition = FALSE;
         }
      }
      
      if (CheckNextCondition == TRUE)
      {
         if (m_HeadArray.Cells[Head].UnpairedComplementCount > 0)
         {
            if ((m_HeadArray.SequencePosition - m_HeadArray.Cells[Head].LastComplementIndex) < PARMS_GetUnpairedDistanceThreshold())
            {
               m_HeadArray.Cells[Head].UnpairedComplementCount += 1;
            }
         }
         else
         {
            m_HeadArray.Cells[Head].UnpairedComplementCount += 1;
         }
      }
      
      m_HeadArray.Cells[Head].LastComplementIndex = m_HeadArray.SequencePosition;
   }
   else
   {
      CheckNextCondition = TRUE;
      if (m_HeadArray.Cells[Head].UnpairedComplementCount > 0)
      {
         if ((m_HeadArray.SequencePosition - m_HeadArray.Cells[Head].LastComplementIndex) < PARMS_GetPairedDistanceThreshold())
         {
            m_HeadArray.Cells[Head].PairedRepeatCount += 1;
            m_HeadArray.Cells[Head].UnpairedComplementCount -= 1;
            CheckNextCondition = FALSE;
         }
      }
      
      if (CheckNextCondition == TRUE)
      {
         if (m_HeadArray.Cells[Head].UnpairedRepeatCount > 0)
         {
            if ((m_HeadArray.SequencePosition - m_HeadArray.Cells[Head].LastHeadRepeatIndex) < PARMS_GetUnpairedDistanceThreshold())
            {
               m_HeadArray.Cells[Head].UnpairedRepeatCount += 1;
            }
         }
         else
         {
            m_HeadArray.Cells[Head].UnpairedRepeatCount += 1;
         }
      }
      
      m_HeadArray.Cells[Head].LastHeadRepeatIndex = m_HeadArray.SequencePosition;
   }

   m_HeadArray.SequencePosition += InvertedDepth;
}

/*********************************************************************************/

void DIST_SetPiecesArrayElement(CNFADT_HEAD_TYPE Head, 
                                 ULONG InvertedDepth)
{
   ULONG i;

   for (i = 0; i < (InvertedDepth + 1); i++)
   {
      if (m_PiecesArray.SequencePosition < m_PiecesArray.Length)
      {
         if ((InvertedDepth + 1) >= PARMS_GetMinimumStructuralLength())
         {
            m_PiecesArray.PieceHead[m_PiecesArray.SequencePosition] = Head;
         }
         else
         {
            m_PiecesArray.PieceHead[m_PiecesArray.SequencePosition] = 0;
         }
         
         m_PiecesArray.SequencePosition += 1;
      }
      else
      {
         printf("Error: trying to set an index longer than array in DIST_SetPiecesArrayElement.\n");
         exit(1);
      }
   }
}

/*********************************************************************************/

void DIST_ScanForStructuralPiece(CNFADT_DICTIONARY_TYPE *Dictionary, 
                                 CNFADT_HEAD_TYPE Head,
                                 BOOLEAN ReverseComplement)
{
   CNFADT_NODE_TYPE *Location;
   CNFADT_NODE_TYPE *Ignore;
   
   if (CNFADT_FindRule(Dictionary, Head, &Location, &Ignore) == FALSE)
   {
      printf("Error: unable to scan for structural pieces.\n");
      exit(1);
   }
   else
   {
      if (Location->Entry.Metric.DistanceFromStructuralAddition == 0)
      {
         DIST_SetHeadArrayElement(Head, Location->Entry.Metric.InvertedDepth, ReverseComplement);
         DIST_SetPiecesArrayElement(Head, Location->Entry.Metric.InvertedDepth);
      }
      else
      {
         if (ReverseComplement == FALSE)
         {
            DIST_ScanForStructuralPiece(Dictionary, 
                                          Location->Entry.Rule.Body.Type2Body[0].Entry->Rule.Head, 
                                          Location->Entry.Rule.Body.Type2Body[0].ReverseComplement);
            DIST_ScanForStructuralPiece(Dictionary, 
                                          Location->Entry.Rule.Body.Type2Body[1].Entry->Rule.Head, 
                                          Location->Entry.Rule.Body.Type2Body[1].ReverseComplement);
         }
         else
         {
            DIST_ScanForStructuralPiece(Dictionary, 
                                          Location->Entry.Rule.Body.Type2Body[1].Entry->Rule.Head, 
                                          NOT(Location->Entry.Rule.Body.Type2Body[1].ReverseComplement));
            DIST_ScanForStructuralPiece(Dictionary, 
                                          Location->Entry.Rule.Body.Type2Body[0].Entry->Rule.Head, 
                                          NOT(Location->Entry.Rule.Body.Type2Body[0].ReverseComplement));
         }
      }
   }
}

/*********************************************************************************/

void DIST_ScanHeadArrayForStructuralCoverage(ULONG *TotalUnpairedRepeats,
                                              ULONG *TotalPaired)
{
   ULONG i;
   ULONG CurrentID;
   ULONG UnpairedCount;
   ULONG PairedCount;
   
   *TotalUnpairedRepeats = 0;
   *TotalPaired = 0;

   CurrentID = 0;
   UnpairedCount = 0;
   PairedCount = 0;

   for (i = 0; i < m_HeadArray.Length; i++)
   {
      if (m_HeadArray.Cells[i].InvertedDepth > CurrentID)
      {
         if ((CurrentID + 1) >= PARMS_GetMinimumStructuralLength())
         {
            *TotalUnpairedRepeats += (UnpairedCount * (CurrentID + 1));
            *TotalPaired += (2 * PairedCount * (CurrentID + 1));
         }
         
         CurrentID = m_HeadArray.Cells[i].InvertedDepth;
         UnpairedCount = 0;
         PairedCount = 0;
      }

      if (m_HeadArray.Cells[i].UnpairedRepeatCount > 1)
      {
         UnpairedCount += m_HeadArray.Cells[i].UnpairedRepeatCount;
      }
      if (m_HeadArray.Cells[i].UnpairedComplementCount > 1)
      {
         UnpairedCount += m_HeadArray.Cells[i].UnpairedComplementCount;
      }
      
      PairedCount += m_HeadArray.Cells[i].PairedRepeatCount;
   }

   if ((CurrentID + 1) >= PARMS_GetMinimumStructuralLength())
   {
      *TotalUnpairedRepeats += (UnpairedCount * (CurrentID + 1));
      *TotalPaired += (2 * PairedCount * (CurrentID + 1));
   }
}

/*********************************************************************************/

void DIST_ScanPiecesArrayForStructuralMultiplicity(void)
{
   ULONG i;
   
   for (i = 0; i < m_PiecesArray.Length; i++)
   {
      if (m_PiecesArray.PieceHead[i] < m_HeadArray.Length)
      {
         if ((m_HeadArray.Cells[m_PiecesArray.PieceHead[i]].UnpairedRepeatCount <= 1) &&
             (m_HeadArray.Cells[m_PiecesArray.PieceHead[i]].PairedRepeatCount == 0))
         {
            m_PiecesArray.PieceHead[i] = 0;
         }
      }
      else
      {
         printf("Error: unable to scan for structural multiplicity.\n");
         exit(1);
      }
   }
}
