/*****************************************************************/
/*  ELEC 999                                                     */
/*  University of Nebraska-Lincoln                               */
/*  Department of Electrical Engineering                         */
/*  David J. Russell                                             */
/*  5-26-08                                                      */
/*                                                               */
/* 01/30/09 - Added % distribution to output type, version 1.1   */
/* 01/31/09 - Added windowing across an input, version 1.2       */
/* 02/14/09 - Added html output type where text is colored to    */
/*             identify the structural pieces. Modified          */
/*             VectorADT to allow longer vectors, version 1.5    */
/*****************************************************************/

#define _FILEIO

/*********************************************************************************/
/* Included Header Files */
/*********************************************************************************/
#include "Globals.h"
#include "StringADT.h"
#include "StringListADT.h"
#include "VectorADT.h"
#include "CNFGrammarADT.h"
#include "FileIO.h"
#include "Parameters.h"

/*********************************************************************************/
/* Private Module Constants */
/*********************************************************************************/
#define FILEIO_SEQUENCE_PART_LENGTH                   512

#define FILEIO_FILE_HEADER_LINE                       "--->ICYK Grammar File<---"

#define FILEIO_HTML_BASEPAIRS_PER_LINE                120
#define FILEIO_HTML_COLORS                            7

/*********************************************************************************/
/* Private Type Definitions */
/*********************************************************************************/
typedef void (*FILEIO_POINTER_TO_STATE_FUNCTION) (BYTE);

typedef enum
{
   FILEIO_STATE_IDLE,
   FILEIO_STATE_HEADER_LINE,
   FILEIO_STATE_SEQUENCE,
} FILEIO_STATE;

typedef struct
{
   FILEIO_STATE State;
   FILEIO_POINTER_TO_STATE_FUNCTION StateFunction;
} FILEIO_STATE_ENTRY;

/*********************************************************************************/
/* Private Module Function Prototypes */
/*********************************************************************************/

/* States */
static void FILEIO_StateIdle (BYTE Ch);
static void FILEIO_StateHeaderLine(BYTE Ch);
static void FILEIO_StateSequence(BYTE Ch);

static void FILEIO_StoreSequencePart(void);
static void FILEIO_CreateNewSequence(void);
static void FILEIO_StoreLastSequencePart(void);
static void FILEIO_CreateNewHeaderLine(void);
static void FILEIO_StoreHeaderLine(void);

/*********************************************************************************/
/* Private Module Variables */
/*********************************************************************************/
static STRINGLISTADT_STRINGLIST_TYPE *m_Sequences;
static STRINGLISTADT_STRINGLIST_TYPE *m_MergedAlphabetSequences;
static BYTE m_SequencePart[FILEIO_SEQUENCE_PART_LENGTH];
static BYTE m_MergedAlphabetSequencePart[FILEIO_SEQUENCE_PART_LENGTH];
static USHORT m_CurrentSequenceIndex;
static STRINGADT_STRING_TYPE m_CurrentSequence;
static STRINGADT_STRING_TYPE m_CurrentMergedAlphabetSequence;
static STRINGLISTADT_STRINGLIST_TYPE *m_HeaderLines;
static STRINGADT_STRING_TYPE m_CurrentHeaderLine;
static PARMS_SEQUENCE_TYPE m_SequenceType;

static FILEIO_STATE m_State;

const static FILEIO_STATE_ENTRY m_StateTable[] =
{
   {FILEIO_STATE_IDLE, FILEIO_StateIdle},
   {FILEIO_STATE_HEADER_LINE, FILEIO_StateHeaderLine},
   {FILEIO_STATE_SEQUENCE, FILEIO_StateSequence},
};

/*
   A,S,T,X
   B,D,N
   C
   E,K,Q,R,Z
   F
   G
   H
   I,L,M,V
   P
   W
   Y
*/

static const BYTE m_MergedAlphabetLookup[ALPHABETIC_CHARACTERS] = 
{
   'A',
   'B',
   'C',
   'B',
   'E',
   'F',
   'G',
   'H',
   'I',
   'A',  /* Unknown -- X is unknown A.A. */
   'E',
   'I',
   'I',
   'B',
   'A',  /* Unknown -- X is unknown A.A. */
   'P',
   'E',
   'E',
   'A',
   'A',
   'A',  /* Unknown -- X is unknown A.A. */
   'I',
   'W',
   'A',
   'Y',
   'E',
   'A',  /* Unknown -- X is unknown A.A. */
};

/*********************************************************************************/
/* Public and Private Module Functions */
/*********************************************************************************/

void FILEIO_ReadFASTASequences(STRINGLISTADT_STRINGLIST_TYPE *Sequences,
                                 STRINGLISTADT_STRINGLIST_TYPE *MergedAlphabetSequences,
                                 STRINGLISTADT_STRINGLIST_TYPE *HeaderLines)
{
   FILE *infile;
   BYTE Ch;
   
   m_State = FILEIO_STATE_IDLE;
   m_CurrentSequenceIndex = 0;
   m_Sequences = Sequences;
   m_MergedAlphabetSequences = MergedAlphabetSequences;
   m_HeaderLines = HeaderLines;

   m_SequenceType = PARMS_SEQUENCE_DNA;

   infile = fopen(PARMS_GetInputFileName(), "r");
   if (infile != NULL)
   {
      do
      {
         Ch = getc(infile);
         m_StateTable[m_State].StateFunction(Ch);
      } while (Ch != EOF);

      fclose(infile);
   }
   else
   {
      printf("Cannot open %s.\n", PARMS_GetInputFileName());
      exit(1);
   }
   
   PARMS_SetSequenceType(m_SequenceType);
}

/*********************************************************************************/

void FILEIO_PrintVerboseDictionaryToFile(CNFADT_DICTIONARY_TYPE *Dictionary)
{
   CNFADT_NODE_TYPE *TempPointer;
   FILE *outfile;
   
   TempPointer = Dictionary->List;

   outfile = fopen(PARMS_GetGrammarOutputFileName(), "w");
   if (outfile == NULL)
   {
      printf("Cannot create %s.\n", PARMS_GetGrammarOutputFileName());
      exit(1);
   }

   fprintf(outfile, "%s\n", FILEIO_FILE_HEADER_LINE);   

   fprintf(outfile, "Next Head Label:         %lu\n\n", Dictionary->CurrentHeadLabel);
   
   while (TempPointer != NULL)
   {
      fprintf(outfile, "Type:                    %d\n", TempPointer->Entry.Rule.Type);
      if (TempPointer->Entry.Rule.Type == CNFADT_RULE_TYPE1)
      {
         fprintf(outfile, "Head:                    %lu\n", TempPointer->Entry.Rule.Head);
         fprintf(outfile, "Body:                    %c\n", TempPointer->Entry.Rule.Body.Type1Body.Label);
         fprintf(outfile, "ID:                      %lu\n", TempPointer->Entry.Metric.InvertedDepth);
         fprintf(outfile, "DSA:                     %lu\n", TempPointer->Entry.Metric.DistanceFromStructuralAddition);
         fprintf(outfile, "Both Bodies:             %d\n", TempPointer->Entry.Metric.BothBodiesAreStructuralPieces);
         fprintf(outfile, "Variance:                %lu\n", TempPointer->Entry.Metric.BodyDepthDifference);
         fprintf(outfile, "Repeat Head Count:       %lu\n", TempPointer->Entry.Metric.HeadRepeatCount);
         fprintf(outfile, "Non-overlap Heads:       %lu\n", TempPointer->Entry.Metric.NonOverlappingHeadRepeatCount);
         fprintf(outfile, "Structural Vector:       ");
         VECTORADT_PrintVector(&(TempPointer->Entry.Metric.StructuralComponents), outfile);
         fprintf(outfile, "\nComplement Vector:       ");
         VECTORADT_PrintVector(&(TempPointer->Entry.Metric.ComplementComponents), outfile);
         fprintf(outfile, "\nSelfComplement Vector:   ");
         VECTORADT_PrintVector(&(TempPointer->Entry.Metric.SelfComplementComponents), outfile);
         fprintf(outfile, "\nSum Vector:              ");
         VECTORADT_PrintVector(&(TempPointer->Entry.Metric.SumOfComponents), outfile);
         fprintf(outfile, "\nDiff Vector:             ");
         VECTORADT_PrintVector(&(TempPointer->Entry.Metric.AbsDifferenceOfComponents), outfile);
         fprintf(outfile, "\n\n");
      }
      else
      {
         fprintf(outfile, "Head:                    %lu\n", TempPointer->Entry.Rule.Head);
         fprintf(outfile, "Self Reverse Complement: %d\n", TempPointer->Entry.Metric.SelfReverseComplement);
         fprintf(outfile, "Body0:                   %lu\n", TempPointer->Entry.Rule.Body.Type2Body[0].Entry->Rule.Head);
         fprintf(outfile, "Self Reverse Complement: %d\n", TempPointer->Entry.Rule.Body.Type2Body[0].Entry->Metric.SelfReverseComplement);
         fprintf(outfile, "Reverse Complement:      %d\n", TempPointer->Entry.Rule.Body.Type2Body[0].ReverseComplement);
         fprintf(outfile, "Body1:                   %lu\n", TempPointer->Entry.Rule.Body.Type2Body[1].Entry->Rule.Head);
         fprintf(outfile, "Self Reverse Complement: %d\n", TempPointer->Entry.Rule.Body.Type2Body[1].Entry->Metric.SelfReverseComplement);
         fprintf(outfile, "Reverse Complement:      %d\n", TempPointer->Entry.Rule.Body.Type2Body[1].ReverseComplement);
         fprintf(outfile, "ID:                      %lu\n", TempPointer->Entry.Metric.InvertedDepth);
         fprintf(outfile, "DSA:                     %lu\n", TempPointer->Entry.Metric.DistanceFromStructuralAddition);
         fprintf(outfile, "Both Bodies:             %d\n", TempPointer->Entry.Metric.BothBodiesAreStructuralPieces);
         fprintf(outfile, "Variance:                %lu\n", TempPointer->Entry.Metric.BodyDepthDifference);
         fprintf(outfile, "Repeat Head Count:       %lu\n", TempPointer->Entry.Metric.HeadRepeatCount);
         fprintf(outfile, "Non-overlap Heads:       %lu\n", TempPointer->Entry.Metric.NonOverlappingHeadRepeatCount);
         fprintf(outfile, "Structural Vector:       ");
         VECTORADT_PrintVector(&(TempPointer->Entry.Metric.StructuralComponents), outfile);
         fprintf(outfile, "\nComplement Vector:       ");
         VECTORADT_PrintVector(&(TempPointer->Entry.Metric.ComplementComponents), outfile);
         fprintf(outfile, "\nSelfComplement Vector:   ");
         VECTORADT_PrintVector(&(TempPointer->Entry.Metric.SelfComplementComponents), outfile);
         fprintf(outfile, "\nSum Vector:              ");
         VECTORADT_PrintVector(&(TempPointer->Entry.Metric.SumOfComponents), outfile);
         fprintf(outfile, "\nDiff Vector:             ");
         VECTORADT_PrintVector(&(TempPointer->Entry.Metric.AbsDifferenceOfComponents), outfile);
         fprintf(outfile, "\n\n");
      }
      TempPointer = TempPointer->Next;
   }
   
   fclose(outfile);
}

/*********************************************************************************/

void FILEIO_PrintCompressedDictionaryToFile(CNFADT_DICTIONARY_TYPE *Dictionary)
{
   CNFADT_NODE_TYPE *TempPointer;
   FILE *outfile;
   
   TempPointer = Dictionary->List;

   outfile = fopen(PARMS_GetGrammarOutputFileName(), "w");
   if (outfile == NULL)
   {
      printf("Cannot create %s.\n", PARMS_GetGrammarOutputFileName());
      exit(1);
   }

   fprintf(outfile, "%s\n", FILEIO_FILE_HEADER_LINE);   

   fprintf(outfile, "%lu\n", Dictionary->CurrentHeadLabel);
   
   while (TempPointer != NULL)
   {
      fprintf(outfile, "%d,", TempPointer->Entry.Rule.Type);
      if (TempPointer->Entry.Rule.Type == CNFADT_RULE_TYPE1)
      {
         fprintf(outfile, "%lu,", TempPointer->Entry.Rule.Head);
         fprintf(outfile, "%c,", TempPointer->Entry.Rule.Body.Type1Body.Label);
         fprintf(outfile, "%lu,", TempPointer->Entry.Metric.InvertedDepth);
         fprintf(outfile, "%lu,", TempPointer->Entry.Metric.DistanceFromStructuralAddition);
         fprintf(outfile, "%d,", TempPointer->Entry.Metric.BothBodiesAreStructuralPieces);
         fprintf(outfile, "%lu,", TempPointer->Entry.Metric.BodyDepthDifference);
         fprintf(outfile, "%lu,", TempPointer->Entry.Metric.HeadRepeatCount);
         fprintf(outfile, "%lu,", TempPointer->Entry.Metric.NonOverlappingHeadRepeatCount);
         VECTORADT_PrintVector(&(TempPointer->Entry.Metric.StructuralComponents), outfile);
         fprintf(outfile, ",");
         VECTORADT_PrintVector(&(TempPointer->Entry.Metric.ComplementComponents), outfile);
         fprintf(outfile, ",");
         VECTORADT_PrintVector(&(TempPointer->Entry.Metric.SelfComplementComponents), outfile);
         fprintf(outfile, ",");
         VECTORADT_PrintVector(&(TempPointer->Entry.Metric.SumOfComponents), outfile);
         fprintf(outfile, ",");
         VECTORADT_PrintVector(&(TempPointer->Entry.Metric.AbsDifferenceOfComponents), outfile);
         fprintf(outfile, "\n");
      }
      else
      {
         fprintf(outfile, "%lu,", TempPointer->Entry.Rule.Head);
         fprintf(outfile, "%d,", TempPointer->Entry.Metric.SelfReverseComplement);
         fprintf(outfile, "%lu,", TempPointer->Entry.Rule.Body.Type2Body[0].Entry->Rule.Head);
         fprintf(outfile, "%d,", TempPointer->Entry.Rule.Body.Type2Body[0].Entry->Metric.SelfReverseComplement);
         fprintf(outfile, "%d,", TempPointer->Entry.Rule.Body.Type2Body[0].ReverseComplement);
         fprintf(outfile, "%lu,", TempPointer->Entry.Rule.Body.Type2Body[1].Entry->Rule.Head);
         fprintf(outfile, "%d,", TempPointer->Entry.Rule.Body.Type2Body[1].Entry->Metric.SelfReverseComplement);
         fprintf(outfile, "%d,", TempPointer->Entry.Rule.Body.Type2Body[1].ReverseComplement);
         fprintf(outfile, "%lu,", TempPointer->Entry.Metric.InvertedDepth);
         fprintf(outfile, "%lu,", TempPointer->Entry.Metric.DistanceFromStructuralAddition);
         fprintf(outfile, "%d,", TempPointer->Entry.Metric.BothBodiesAreStructuralPieces);
         fprintf(outfile, "%lu,", TempPointer->Entry.Metric.BodyDepthDifference);
         fprintf(outfile, "%lu,", TempPointer->Entry.Metric.HeadRepeatCount);
         fprintf(outfile, "%lu,", TempPointer->Entry.Metric.NonOverlappingHeadRepeatCount);
         VECTORADT_PrintVector(&(TempPointer->Entry.Metric.StructuralComponents), outfile);
         fprintf(outfile, ",");
         VECTORADT_PrintVector(&(TempPointer->Entry.Metric.ComplementComponents), outfile);
         fprintf(outfile, ",");
         VECTORADT_PrintVector(&(TempPointer->Entry.Metric.SelfComplementComponents), outfile);
         fprintf(outfile, ",");
         VECTORADT_PrintVector(&(TempPointer->Entry.Metric.SumOfComponents), outfile);
         fprintf(outfile, ",");
         VECTORADT_PrintVector(&(TempPointer->Entry.Metric.AbsDifferenceOfComponents), outfile);
         fprintf(outfile, "\n");
      }
      TempPointer = TempPointer->Next;
   }
   
   fclose(outfile);
}

/*********************************************************************************/

void FILEIO_PrintDistributionStringToFile(STRINGADT_STRING_TYPE *DistributionString)
{
   FILE *outfile;
   ULONG i;
   
   outfile = fopen(PARMS_GetDistributionOutputFileName(), "w");
   if (outfile == NULL)
   {
      printf("Cannot create %s.\n", PARMS_GetDistributionOutputFileName());
      exit(1);
   }

   for (i = 0; i < DistributionString->Length; i++)
   {
      fprintf(outfile, "%c", DistributionString->Characters[i]);
   }
   
   fclose(outfile);
}

/*********************************************************************************/

void FILEIO_WriteHTMLColoredSequence(STRINGADT_STRING_TYPE *Sequence,
                                     ULONG *SequencePieces,
                                     ULONG Length)
{
   FILE *outfile;
   ULONG j;
   BOOLEAN LastCharacterWasAReturn = FALSE;
   UBYTE Color;
   
   outfile = fopen(PARMS_GetHTMLOutputFileName(), "w");
   if (outfile != NULL)
   {
      fprintf(outfile, "<html>\n");
      fprintf(outfile, "<head>\n");
      fprintf(outfile, "<title>ICYK: Structurally Colored Seqnece</title>\n");
      fprintf(outfile, "<style type=\"text/css\">\n");
      fprintf(outfile, ".c00 {font-family: courier; font-size: 12px; background-color: #FFFFFF; color: #000000;}\n");
      fprintf(outfile, ".c01 {font-family: courier; font-size: 12px; background-color: #000000; color: #CCCCCC;}\n");
      fprintf(outfile, ".c02 {font-family: courier; font-size: 12px; background-color: #0000FF; color: #CCCCCC;}\n");
      fprintf(outfile, ".c03 {font-family: courier; font-size: 12px; background-color: #00FF00; color: #333333;}\n");
      fprintf(outfile, ".c04 {font-family: courier; font-size: 12px; background-color: #00FFFF; color: #333333;}\n");
      fprintf(outfile, ".c05 {font-family: courier; font-size: 12px; background-color: #FF0000; color: #CCCCCC;}\n");
      fprintf(outfile, ".c06 {font-family: courier; font-size: 12px; background-color: #FF00FF; color: #333333;}\n");
      fprintf(outfile, ".c07 {font-family: courier; font-size: 12px; background-color: #FFFF00; color: #333333;}\n");
      fprintf(outfile, "</style>\n");
      fprintf(outfile, "</head>\n");
      fprintf(outfile, "<body>\n");
      fprintf(outfile, "<br>\n");

      fprintf(outfile, "Structurally colored sequence:<br><br>\n");

      for (j = 0; j < Length; j++)
      {
         if (SequencePieces[j] == 0)
         {
            Color = 00;
         }
         else
         {
            Color = (SequencePieces[j] % FILEIO_HTML_COLORS) + 1;
         }
         
         fprintf(outfile, "<span class=\"c%2.2d\">%c</span>", Color, Sequence->Characters[j]);

         LastCharacterWasAReturn = FALSE;
         if (((j + 1) % FILEIO_HTML_BASEPAIRS_PER_LINE) == 0)
         {
            fprintf(outfile, "<br>\n");
            LastCharacterWasAReturn = TRUE;
         }
      }
      
      if (LastCharacterWasAReturn == FALSE)
      {
         fprintf(outfile, "<br>\n");
      }

      fprintf(outfile, "</body>\n");
      fprintf(outfile, "</html>\n");
      fclose(outfile);
   }
   else
   {
      printf("Cannot create %s.\n", PARMS_GetHTMLOutputFileName());
      exit(1);
   }
}

/*********************************************************************************/

void FILEIO_StateIdle(BYTE Ch)
{
   if (Ch == '>')
   {
      FILEIO_CreateNewHeaderLine();
      
      m_State = FILEIO_STATE_HEADER_LINE;
   }
}

/*********************************************************************************/

void FILEIO_StateHeaderLine(BYTE Ch)
{
   if ((Ch == '\n') || (Ch == '\r'))
   {
      FILEIO_StoreHeaderLine();
      FILEIO_CreateNewSequence();

      m_State = FILEIO_STATE_SEQUENCE;
   }
   else
   {
      if (Ch == ':')
      {
         STRINGADT_AppendStringFromChar(&m_CurrentHeaderLine, 
                                          '_', 
                                          "FILEIO_StateHeaderLine");
      }
      else
      {
         STRINGADT_AppendStringFromChar(&m_CurrentHeaderLine, 
                                          Ch, 
                                          "FILEIO_StateHeaderLine");
      }
   }
}

/*********************************************************************************/

void FILEIO_StateSequence(BYTE Ch)
{
   BYTE CH = toupper(Ch);
   
   switch(CH)
   {
      case '>':
         FILEIO_StoreLastSequencePart();
         FILEIO_CreateNewHeaderLine();
         m_State = FILEIO_STATE_HEADER_LINE;
         break;

      case 'A':
      case 'C':
      case 'G':
      case 'T':
      case 'X':
         m_MergedAlphabetSequencePart[m_CurrentSequenceIndex] = m_MergedAlphabetLookup[ALPHA_INDEX(CH)];
         m_SequencePart[m_CurrentSequenceIndex++] = CH;
         if (m_CurrentSequenceIndex >= FILEIO_SEQUENCE_PART_LENGTH)
         {
            FILEIO_StoreSequencePart();
         }
         break;

      case 'U':
         m_MergedAlphabetSequencePart[m_CurrentSequenceIndex] = m_MergedAlphabetLookup[ALPHA_INDEX(CH)];
         m_SequencePart[m_CurrentSequenceIndex++] = CH;
         if (m_CurrentSequenceIndex >= FILEIO_SEQUENCE_PART_LENGTH)
         {
            FILEIO_StoreSequencePart();
         }
         m_SequenceType = PARMS_SEQUENCE_RNA;
         break;

      case 'B':
      case 'D':
      case 'E':
      case 'F':
      case 'H':
      case 'I':
      case 'K':
      case 'L':
      case 'M':
      case 'N':
      case 'P':
      case 'Q':
      case 'R':
      case 'S':
      case 'V':
      case 'W':
      case 'Y':
      case 'Z':
         m_MergedAlphabetSequencePart[m_CurrentSequenceIndex] = m_MergedAlphabetLookup[ALPHA_INDEX(CH)];
         m_SequencePart[m_CurrentSequenceIndex++] = CH;
         if (m_CurrentSequenceIndex >= FILEIO_SEQUENCE_PART_LENGTH)
         {
            FILEIO_StoreSequencePart();
         }
         m_SequenceType = PARMS_SEQUENCE_AMINO_ACID;
         break;

      case EOF:
         FILEIO_StoreLastSequencePart();
         break;

      default:
         break;
   }
}

/*********************************************************************************/

void FILEIO_StoreSequencePart(void)
{
   if (m_CurrentSequenceIndex > 0)
   {
      STRINGADT_AppendStringFromChars(&m_CurrentSequence, 
                                       (UBYTE *) m_SequencePart,
                                       m_CurrentSequenceIndex, 
                                       "FILEIO_StoreSequencePart");
      STRINGADT_AppendStringFromChars(&m_CurrentMergedAlphabetSequence, 
                                       (UBYTE *) m_MergedAlphabetSequencePart,
                                       m_CurrentSequenceIndex, 
                                       "FILEIO_StoreSequencePart");
      m_CurrentSequenceIndex = 0;
   }
}

/*********************************************************************************/

void FILEIO_CreateNewSequence(void)
{
   m_CurrentSequenceIndex = 0;
   STRINGADT_InitializeString(&m_CurrentSequence);
   STRINGADT_InitializeString(&m_CurrentMergedAlphabetSequence);
}

/*********************************************************************************/

void FILEIO_StoreLastSequencePart(void)
{
   FILEIO_StoreSequencePart();
   STRINGLISTADT_AppendStringToStringList(m_Sequences, 
                                          &m_CurrentSequence, 
                                          "FILEIO_StoreLastSequencePart");
   STRINGLISTADT_AppendStringToStringList(m_MergedAlphabetSequences, 
                                          &m_CurrentMergedAlphabetSequence, 
                                          "FILEIO_StoreLastSequencePart");
   STRINGADT_CleanUpString(&m_CurrentSequence);
   STRINGADT_CleanUpString(&m_CurrentMergedAlphabetSequence);
}

/*********************************************************************************/

void FILEIO_CreateNewHeaderLine(void)
{
   STRINGADT_InitializeString(&m_CurrentHeaderLine);
}

/*********************************************************************************/

void FILEIO_StoreHeaderLine(void)
{
   STRINGLISTADT_AppendStringToStringList(m_HeaderLines, 
                                          &m_CurrentHeaderLine, 
                                          "FILEIO_StoreHeaderLine");
   STRINGADT_CleanUpString(&m_CurrentHeaderLine);
}
