/*****************************************************************/
/*  ELEC 999                                                     */
/*  University of Nebraska-Lincoln                               */
/*  Department of Electrical Engineering                         */
/*  David J. Russell                                             */
/*  5-17-08                                                      */
/*                                                               */
/* 01/30/09 - Added % distribution to output type, version 1.1   */
/* 01/31/09 - Added windowing across an input, version 1.2       */
/* 02/03/09 - Added unpaired repeats and modified paired counts, */
/*             version 1.3                                       */
/* 02/07/09 - Added extreme-value % removal for the stats output,*/
/*             version 1.4                                       */
/* 02/14/09 - Added html output type where text is colored to    */
/*             identify the structural pieces. Modified          */
/*             VectorADT to allow longer vectors, version 1.5    */
/* 02/16/09 - Added minimum ID parameter for labeling nodes on   */
/*             .eps outputs, version 1.6                         */
/* 02/17/09 - Added a parameter to control how many nodes below  */
/*             the root of a structural node are shown on the    */
/*             .eps outputs, version 1.7                         */
/*****************************************************************/

#define _ICYK

/*********************************************************************************/
/* Included Header Files */
/*********************************************************************************/
#include "Globals.h"
#include "ICYK.h"
#include "StringADT.h"
#include "StringListADT.h"
#include "VectorADT.h"
#include "CNFGrammarADT.h"
#include "Distribution.h"
#include "Inference.h"
#include "FileIO.h"
#include "PostScript.h"
#include "PostScriptTree.h"
#include "Parameters.h"

/*********************************************************************************/
/* Private Module Constants */
/*********************************************************************************/

#define ICYK_MAX_STRING_FRAGMENT_LENGTH   80
#define ICYK_PERCENTAGE_ARRAY_SIZE        101

/*********************************************************************************/
/* Private Type Definitions */
/*********************************************************************************/

/*********************************************************************************/
/* Private Module Function Prototypes */
/*********************************************************************************/
static void ICYK_SinglePassProcess(STRINGADT_STRING_TYPE *Sequence);
static void ICYK_MultiplePassProcess(STRINGADT_STRING_TYPE *Sequence);
static ULONG ICYK_TrimExtremeValues(ULONG *PercentageArray,
                                    FLOAT *AccumulatedPercentage,
                                    ULONG TotalInArray);
static void ICYK_GenerateTreeString(CNFADT_DICTIONARY_TYPE *Dictionary, 
                                    CNFADT_HEAD_TYPE Head,
                                    BOOLEAN ReverseComplement,
                                    STRINGADT_STRING_TYPE *String,
                                    BOOLEAN RootOfTree,
                                    ULONG HeadDepth,
                                    ULONG DistanceFromStructuralAdditionCount);

/*********************************************************************************/
/* Private Module Variables */
/*********************************************************************************/

/*********************************************************************************/
/* Public and Private Module Functions */
/*********************************************************************************/

INT main(INT argc, BYTE **argv)
{
   STRINGLISTADT_STRINGLIST_TYPE Sequences;
   STRINGLISTADT_STRINGLIST_TYPE Dump1;
   STRINGLISTADT_STRINGLIST_TYPE Dump2;

   if (PARMS_ProcessCommandLine(argc, argv) == TRUE)
   {
      STRINGLISTADT_InitializeStringList(&Sequences);
      STRINGLISTADT_InitializeStringList(&Dump1);
      STRINGLISTADT_InitializeStringList(&Dump2);

      srand((int) time(NULL));
      
      FILEIO_ReadFASTASequences(&Sequences, &Dump1, &Dump2);

      switch (PARMS_GetOutputType())
      {
         case PARMS_OUTPUT_VERBOSE_GRAMMAR_ONLY:
         case PARMS_OUTPUT_COMPRESSED_GRAMMAR_ONLY:
         case PARMS_OUTPUT_VERBOSE_GRAMMAR_AND_TREE:
         case PARMS_OUTPUT_COMPRESSED_GRAMMAR_AND_TREE:
         case PARMS_OUTPUT_HTML_COLORED_SEQUENCE:
            ICYK_SinglePassProcess(&(Sequences.Strings[0]));
            break;
            
         case PARMS_OUTPUT_STRUCTURAL_DISTRIBUTION_PERCENTAGE:
            ICYK_MultiplePassProcess(&(Sequences.Strings[0]));
            break;            
      }
      
      PARMS_Printf("Done.\n", TRUE);

      STRINGLISTADT_CleanUpStringList(&Sequences);
      STRINGLISTADT_CleanUpStringList(&Dump1);
      STRINGLISTADT_CleanUpStringList(&Dump2);
   }
   
   return(0);
}

/*********************************************************************************/

void ICYK_SinglePassProcess(STRINGADT_STRING_TYPE *Sequence)
{
   CNFADT_HEAD_TYPE Head;
   ULONG HeadDepth;
   STRINGADT_STRING_TYPE TreeString;
   CNFADT_DICTIONARY_TYPE Dictionary;
   ULONG WindowLength;
   FLOAT Dump1;
   FLOAT Dump2;
   ULONG *PiecesArray;
   ULONG PiecesArrayLength;

   if (PARMS_GetDisableWindow() == TRUE)
   {
      WindowLength = Sequence->Length;
   }
   else
   {
      WindowLength = MIN(PARMS_GetWindowLength(), Sequence->Length);
   }

   STRINGADT_InitializeString(&TreeString);
   CNFADT_CreateDictionary(&Dictionary);

   PARMS_Printf("Inferring a CNF grammar.\n", TRUE);
   GI_PerformGrammarInference(&Dictionary, Sequence, 0, WindowLength - 1, TRUE);
   GI_GetRootOfTree(&Head, &HeadDepth);
   GI_CleanUp();

   switch (PARMS_GetOutputType())
   {
      case PARMS_OUTPUT_VERBOSE_GRAMMAR_ONLY:
      case PARMS_OUTPUT_COMPRESSED_GRAMMAR_ONLY:
         break;
         
      case PARMS_OUTPUT_VERBOSE_GRAMMAR_AND_TREE:
      case PARMS_OUTPUT_COMPRESSED_GRAMMAR_AND_TREE:
         PARMS_Printf("Generating PostScript image of tree.\n", TRUE);
         ICYK_GenerateTreeString(&Dictionary, Head, FALSE, &TreeString, TRUE, HeadDepth, 0);
         PST_ParseTreeStringIntoPostScript (&TreeString, HeadDepth + 1);
         break;
         
      case PARMS_OUTPUT_STRUCTURAL_DISTRIBUTION_PERCENTAGE:
      case PARMS_OUTPUT_HTML_COLORED_SEQUENCE:
         break;            
   }

   switch (PARMS_GetOutputType())
   {
      case PARMS_OUTPUT_VERBOSE_GRAMMAR_ONLY:
      case PARMS_OUTPUT_VERBOSE_GRAMMAR_AND_TREE:
         PARMS_Printf("Writing grammar to file.\n", TRUE);
         FILEIO_PrintVerboseDictionaryToFile(&Dictionary);
         break;
         
      case PARMS_OUTPUT_COMPRESSED_GRAMMAR_ONLY:
      case PARMS_OUTPUT_COMPRESSED_GRAMMAR_AND_TREE:
         PARMS_Printf("Writing grammar to file.\n", TRUE);
         FILEIO_PrintCompressedDictionaryToFile(&Dictionary);
         break;
         
      case PARMS_OUTPUT_STRUCTURAL_DISTRIBUTION_PERCENTAGE:
         break;            

      case PARMS_OUTPUT_HTML_COLORED_SEQUENCE:
         DIST_GenerateStructureDistribution(&Dictionary, Head, WindowLength, &Dump1, &Dump2);
         PiecesArrayLength = DIST_GetPiecesArray(&PiecesArray);
         FILEIO_WriteHTMLColoredSequence(Sequence, PiecesArray, PiecesArrayLength);
         DIST_CleanUpPiecesArray();
         break;            
   }

   STRINGADT_CleanUpString(&TreeString);
   CNFADT_DeleteDictionary(&Dictionary);
}

/*********************************************************************************/

void ICYK_MultiplePassProcess(STRINGADT_STRING_TYPE *Sequence)
{
   CNFADT_HEAD_TYPE Head;
   ULONG HeadDepth;
   CNFADT_DICTIONARY_TYPE Dictionary;
   ULONG i;
   ULONG FinalStartingPosition;
   ULONG WindowLength;
   BYTE Fragment[ICYK_MAX_STRING_FRAGMENT_LENGTH];
   STRINGADT_STRING_TYPE DistributionString;
   FLOAT SingleUnpairedRepeatPercentage;
   FLOAT SinglePairedPercentage;
   FLOAT AccumulatedUnpairedRepeatPercentage = 0.0;
   FLOAT AccumulatedPairedPercentage = 0.0;
   ULONG UnpairedRepeatPercentages[ICYK_PERCENTAGE_ARRAY_SIZE];
   ULONG PairedPercentages[ICYK_PERCENTAGE_ARRAY_SIZE];
   ULONG TotalTrimmed;
   
   memset (UnpairedRepeatPercentages, 0, sizeof(ULONG) * ICYK_PERCENTAGE_ARRAY_SIZE);
   memset (PairedPercentages, 0, sizeof(ULONG) * ICYK_PERCENTAGE_ARRAY_SIZE);
   
   if (PARMS_GetDisableWindow() == TRUE)
   {
      WindowLength = Sequence->Length;
   }
   else
   {
      WindowLength = MIN(PARMS_GetWindowLength(), Sequence->Length);
   }
   FinalStartingPosition = Sequence->Length - WindowLength;

   STRINGADT_InitializeString(&DistributionString);

   PARMS_Printf("Estimating Structural Distribution.\n", TRUE);
   for (i = 0; i <= FinalStartingPosition; i++)
   {
      CNFADT_CreateDictionary(&Dictionary);
      
      GI_PerformGrammarInference(&Dictionary, Sequence, i, i + WindowLength - 1, FALSE);
      GI_GetRootOfTree(&Head, &HeadDepth);
      GI_CleanUp();

      switch (PARMS_GetOutputType())
      {
         case PARMS_OUTPUT_VERBOSE_GRAMMAR_ONLY:
         case PARMS_OUTPUT_VERBOSE_GRAMMAR_AND_TREE:
         case PARMS_OUTPUT_COMPRESSED_GRAMMAR_ONLY:
         case PARMS_OUTPUT_COMPRESSED_GRAMMAR_AND_TREE:
            break;
         
         case PARMS_OUTPUT_STRUCTURAL_DISTRIBUTION_PERCENTAGE:
            DIST_GenerateStructureDistribution(&Dictionary, 
                                                Head, 
                                                WindowLength,
                                                &SingleUnpairedRepeatPercentage,
                                                &SinglePairedPercentage);
            DIST_CleanUpPiecesArray();
            
            AccumulatedUnpairedRepeatPercentage += SingleUnpairedRepeatPercentage;
            AccumulatedPairedPercentage += SinglePairedPercentage;
            UnpairedRepeatPercentages[(ULONG) SingleUnpairedRepeatPercentage] += 1;
            PairedPercentages[(ULONG) SinglePairedPercentage] += 1;
            
            sprintf (Fragment, "%4lu %5.2f %5.2f\n", i, SingleUnpairedRepeatPercentage, SinglePairedPercentage);
            STRINGADT_AppendStringFromChars(&DistributionString, (UBYTE *) Fragment, strlen(Fragment), "ICYK_MultiplePassProcess");
            break;            

         case PARMS_OUTPUT_HTML_COLORED_SEQUENCE:
            break;            
      }
      
      CNFADT_DeleteDictionary(&Dictionary);
      
      PARMS_PrintfProgress(i, FinalStartingPosition, TRUE);
   }

   TotalTrimmed = ICYK_TrimExtremeValues(&(UnpairedRepeatPercentages[0]), &AccumulatedUnpairedRepeatPercentage, (FinalStartingPosition + 1));
   ICYK_TrimExtremeValues(&(PairedPercentages[0]), &AccumulatedPairedPercentage, (FinalStartingPosition + 1));

   sprintf (Fragment, "Average unpaired repeat: \t%5.2f %%%%\n", (AccumulatedUnpairedRepeatPercentage / ((FLOAT) (FinalStartingPosition + 1 - TotalTrimmed))));
   PARMS_Printf(Fragment, TRUE);
   sprintf (Fragment, "Average pairing: \t\t%5.2f %%%%\n", (AccumulatedPairedPercentage / ((FLOAT) (FinalStartingPosition + 1 - TotalTrimmed))));
   PARMS_Printf(Fragment, TRUE);
   FILEIO_PrintDistributionStringToFile(&DistributionString);
   STRINGADT_CleanUpString(&DistributionString);
}

/*********************************************************************************/

ULONG ICYK_TrimExtremeValues(ULONG *PercentageArray,
                              FLOAT *AccumulatedPercentage,
                              ULONG TotalInArray)
{
   ULONG TotalAmountTrimmed = 0;
   ULONG Count_Max;
   ULONG Count_Min;
   ULONG i_Max = ICYK_PERCENTAGE_ARRAY_SIZE - 1;
   ULONG i_Min = 0;
   
   Count_Max = ((TotalInArray * ((ULONG) PARMS_GetExtremeValueRemovalPercentage())) / 100);
   Count_Min = ((TotalInArray * ((ULONG) PARMS_GetExtremeValueRemovalPercentage())) / 100);

   while ((Count_Max > 0) && (Count_Min > 0) && (i_Max > 0) && (i_Min < ICYK_PERCENTAGE_ARRAY_SIZE))
   {
      if (Count_Max > 0)
      {
         if (PercentageArray[i_Max] > 0)
         {
            PercentageArray[i_Max] -= 1;
            *AccumulatedPercentage -= i_Max;
            Count_Max--;
            TotalAmountTrimmed++;
         }
         else
         {
            i_Max--;
         }
      }
      
      if (Count_Min > 0)
      {
         if (PercentageArray[i_Min] > 0)
         {
            PercentageArray[i_Min] -= 1;
            *AccumulatedPercentage -= i_Min;
            Count_Min--;
            TotalAmountTrimmed++;
         }
         else
         {
            i_Min++;
         }
      }
   }

   return (TotalAmountTrimmed);
}

/*********************************************************************************/

void ICYK_GenerateTreeString(CNFADT_DICTIONARY_TYPE *Dictionary, 
                              CNFADT_HEAD_TYPE Head,
                              BOOLEAN ReverseComplement,
                              STRINGADT_STRING_TYPE *String,
                              BOOLEAN RootOfTree,
                              ULONG HeadDepth,
                              ULONG DistanceFromStructuralAdditionCount)
{
   CNFADT_NODE_TYPE *Location;
   CNFADT_NODE_TYPE *Ignore;
   BYTE Fragment[ICYK_MAX_STRING_FRAGMENT_LENGTH];
   
   if (CNFADT_FindRule(Dictionary, Head, &Location, &Ignore) == FALSE)
   {
      printf("Error: unable to generate tree string.\n");
      exit(1);
   }
   else
   {

      if (Location->Entry.Metric.DistanceFromStructuralAddition == 0)
      {
         DistanceFromStructuralAdditionCount++;
      }

      if ((Location->Entry.Rule.Type == CNFADT_RULE_TYPE1) || 
          ((PARMS_GetEnableSelectiveNodeOutput() == TRUE) && 
           (DistanceFromStructuralAdditionCount > PARMS_GetSelectiveNodeOutputPathDistance())))
      {
         if (Location->Entry.Metric.InvertedDepth < PARMS_GetMinimumIDForLabel())
         {
            sprintf (Fragment, ":%lu", (HeadDepth - Location->Entry.Metric.InvertedDepth));
         }
         else
         {
            if (ReverseComplement == FALSE)
            {
               sprintf (Fragment, "\"%lu\":%lu", Location->Entry.Rule.Head, (HeadDepth - Location->Entry.Metric.InvertedDepth));
            }
            else
            {
               sprintf (Fragment, "\"%lu'\":%lu", Location->Entry.Rule.Head, (HeadDepth - Location->Entry.Metric.InvertedDepth));
            }
         }
         STRINGADT_AppendStringFromChars(String, (UBYTE *) Fragment, strlen(Fragment), "ICYK_GenerateTreeString");
      }
      else
      {
         sprintf (Fragment, "(");
         STRINGADT_AppendStringFromChars(String, (UBYTE *) Fragment, strlen(Fragment), "ICYK_GenerateTreeString");

         if (ReverseComplement == FALSE)
         {
            ICYK_GenerateTreeString(Dictionary, Location->Entry.Rule.Body.Type2Body[0].Entry->Rule.Head, Location->Entry.Rule.Body.Type2Body[0].ReverseComplement, String, FALSE, Location->Entry.Metric.InvertedDepth, DistanceFromStructuralAdditionCount);
            sprintf (Fragment, ",");
            STRINGADT_AppendStringFromChars(String, (UBYTE *) Fragment, strlen(Fragment), "ICYK_GenerateTreeString");
            ICYK_GenerateTreeString(Dictionary, Location->Entry.Rule.Body.Type2Body[1].Entry->Rule.Head, Location->Entry.Rule.Body.Type2Body[1].ReverseComplement, String, FALSE, Location->Entry.Metric.InvertedDepth, DistanceFromStructuralAdditionCount);
            if (RootOfTree == TRUE)
            {
               sprintf (Fragment, ");");
            }
            else
            {
               if (Location->Entry.Metric.InvertedDepth < PARMS_GetMinimumIDForLabel())
               {
                  sprintf (Fragment, "):%lu", (HeadDepth - Location->Entry.Metric.InvertedDepth));
               }
               else
               {
                  if (Location->Entry.Metric.SelfReverseComplement == TRUE)
                  {
                     sprintf (Fragment, ")\"%lu*\":%lu", Location->Entry.Rule.Head, (HeadDepth - Location->Entry.Metric.InvertedDepth));
                  }
                  else
                  {
                     sprintf (Fragment, ")\"%lu\":%lu", Location->Entry.Rule.Head, (HeadDepth - Location->Entry.Metric.InvertedDepth));
                  }
               }
            }
         }
         else
         {
            ICYK_GenerateTreeString(Dictionary, Location->Entry.Rule.Body.Type2Body[1].Entry->Rule.Head, NOT(Location->Entry.Rule.Body.Type2Body[1].ReverseComplement), String, FALSE, Location->Entry.Metric.InvertedDepth, DistanceFromStructuralAdditionCount);
            sprintf (Fragment, ",");
            STRINGADT_AppendStringFromChars(String, (UBYTE *) Fragment, strlen(Fragment), "ICYK_GenerateTreeString");
            ICYK_GenerateTreeString(Dictionary, Location->Entry.Rule.Body.Type2Body[0].Entry->Rule.Head, NOT(Location->Entry.Rule.Body.Type2Body[0].ReverseComplement), String, FALSE, Location->Entry.Metric.InvertedDepth, DistanceFromStructuralAdditionCount);
            if (RootOfTree == TRUE)
            {
               sprintf (Fragment, ");");
            }
            else
            {
               if (Location->Entry.Metric.InvertedDepth < PARMS_GetMinimumIDForLabel())
               {
                  sprintf (Fragment, "):%lu", (HeadDepth - Location->Entry.Metric.InvertedDepth));
               }
               else
               {
                  if (Location->Entry.Metric.SelfReverseComplement == TRUE)
                  {
                     sprintf (Fragment, ")\"%lu*\":%lu", Location->Entry.Rule.Head, (HeadDepth - Location->Entry.Metric.InvertedDepth));
                  }
                  else
                  {
                     sprintf (Fragment, ")\"%lu'\":%lu", Location->Entry.Rule.Head, (HeadDepth - Location->Entry.Metric.InvertedDepth));
                  }
               }
            }
         }

         STRINGADT_AppendStringFromChars(String, (UBYTE *) Fragment, strlen(Fragment), "ICYK_GenerateTreeString");
      }
   }
}

