/*****************************************************************/
/*  ELEC 999                                                     */
/*  University of Nebraska-Lincoln                               */
/*  Department of Electrical Engineering                         */
/*  David J. Russell                                             */
/*  5-26-08                                                      */
/*                                                               */
/* 01/30/09 - Added % distribution to output type, version 1.1   */
/* 01/31/09 - Added windowing across an input, version 1.2       */
/*****************************************************************/

#define _GI

/*********************************************************************************/
/* Included Header Files */
/*********************************************************************************/
#include "Globals.h"
#include "VectorADT.h"
#include "CNFGrammarADT.h"
#include "StringADT.h"
#include "Inference.h"
#include "Parameters.h"

/*********************************************************************************/
/* Private Module Constants */
/*********************************************************************************/
#define INF_NUMBER_OF_PRELOADED_CHARACTERS_DNA_RNA    2
#define INF_NUMBER_OF_PRELOADED_CHARACTERS_AMINO_ACID 23

static const UBYTE m_DNARNAPreloadedCharacters[INF_NUMBER_OF_PRELOADED_CHARACTERS_DNA_RNA] = 
{  /* So, A = 0, C = 1, G = 1', T/U = 0' */
   'A', 'C'
};

static const UBYTE m_AminoAcidPreloadedCharacters[INF_NUMBER_OF_PRELOADED_CHARACTERS_AMINO_ACID] = 
{
   'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'K', 'L', 'M', 'N', 
   'P', 'Q', 'R', 'S', 'T', 'V', 'W', 'X', 'Y', 'Z'
};

/*********************************************************************************/
/* Private Type Definitions */
/*********************************************************************************/
typedef struct
{
   CNFADT_NODE_TYPE *GrammarNode;
   BOOLEAN ReverseComplement;
   BOOLEAN RuleSet;
} GI_CELL_INFO_TYPE;

typedef struct
{
   GI_CELL_INFO_TYPE *RowOfCells;
   ULONG NumberOfCells;
} GI_ROW_TYPE;

/*********************************************************************************/
/* Private Module Function Prototypes */
/*********************************************************************************/
static void GI_Initialize(ULONG SequenceLength);
static void GI_PlaceRuleInCell(ULONG Depth, ULONG ColumnBegin, GI_CELL_INFO_TYPE RuleInfo);
static void GI_PreloadType1Rules(CNFADT_DICTIONARY_TYPE *Dictionary);
static void GI_CreateType1Rules(CNFADT_DICTIONARY_TYPE *Dictionary,
                                 STRINGADT_STRING_TYPE *Sequence,
                                 ULONG FirstIndex,
                                 ULONG LastIndex);
static void GI_CreateType2Rules(CNFADT_DICTIONARY_TYPE *Dictionary,
                                 ULONG Depth);
static void GI_LoadType2Rule(CNFADT_RULE_AND_METRIC_TYPE *NewEntry,
                              ULONG Row1,
                              ULONG Row2,
                              ULONG Cell1,
                              ULONG Cell2,
                              ULONG Depth);
static void GI_GenerateInitialVectors(CNFADT_RULE_AND_METRIC_TYPE *NewEntry);
static BOOLEAN GI_SelectRuleAOverRuleB(CNFADT_RULE_AND_METRIC_TYPE *RuleAInfo,
                                       CNFADT_RULE_AND_METRIC_TYPE *RuleBInfo);
static void GI_AdjustVectorsViaHeadCountAtThisDepth(ULONG Depth, 
                                                   CNFADT_DICTIONARY_TYPE *Dictionary);
static void GI_DetermineNonOverlappingHeadRepeats(ULONG Depth, 
                                                   CNFADT_DICTIONARY_TYPE *Dictionary);
static void GI_PruneUselessRules(CNFADT_DICTIONARY_TYPE *Dictionary, ULONG Depth, ULONG Cell);

/*********************************************************************************/
/* Private Module Variables */
/*********************************************************************************/
static GI_ROW_TYPE *m_InferenceRows;
static ULONG m_MaxDepth;
static BOOLEAN m_AtLeastOneRepeatAtThisDepth;

/*********************************************************************************/
/* Public and Private Module Functions */
/*********************************************************************************/

void GI_PerformGrammarInference(CNFADT_DICTIONARY_TYPE *Dictionary,
                                 STRINGADT_STRING_TYPE *Sequence,
                                 ULONG FirstIndex,
                                 ULONG LastIndex,
                                 BOOLEAN ShowProgress)
{
   ULONG i;
   
   GI_Initialize(LastIndex - FirstIndex + 1);

   GI_PreloadType1Rules(Dictionary);
   GI_CreateType1Rules(Dictionary, Sequence, FirstIndex, LastIndex);
   GI_DetermineNonOverlappingHeadRepeats(0, Dictionary);

   PARMS_Printf("Creating dictionary.\n", ShowProgress);
   PARMS_PrintfProgress(0, m_MaxDepth, ShowProgress);
   for (i = 1; i < m_MaxDepth; i++)
   {
      GI_CreateType2Rules(Dictionary, i);

      GI_AdjustVectorsViaHeadCountAtThisDepth(i, Dictionary);
      
      PARMS_PrintfProgress(i, m_MaxDepth, ShowProgress);
   }
   PARMS_PrintfProgress(m_MaxDepth, m_MaxDepth, ShowProgress);

   PARMS_Printf("Pruning useless rules.\n", ShowProgress);
   GI_PruneUselessRules(Dictionary, m_MaxDepth - 1, 0);
   
   PARMS_Printf("Relabeling rules.\n", ShowProgress);
   CNFADT_RelabelRulesInDictionary(Dictionary);
}

/*********************************************************************************/

void GI_Initialize(ULONG SequenceLength)
{
   ULONG i;
   ULONG j;
   
   m_MaxDepth = SequenceLength;
   m_AtLeastOneRepeatAtThisDepth = TRUE;
   
   m_InferenceRows = (GI_ROW_TYPE *) malloc(sizeof(GI_ROW_TYPE) * m_MaxDepth);

   if (m_InferenceRows == NULL)
   {
      printf("Error: malloc failed in GI_Initialize.\n");
      exit(1);
   }
   else
   {
      for (i = 0; i < m_MaxDepth; i++)
      {
         m_InferenceRows[i].RowOfCells = (GI_CELL_INFO_TYPE *) malloc(sizeof(GI_CELL_INFO_TYPE) * (m_MaxDepth - i));

         if (m_InferenceRows[i].RowOfCells == NULL)
         {
            printf("Error: malloc failed in GI_Initialize.\n");
            exit(1);
         }
         else
         {
            m_InferenceRows[i].NumberOfCells = (m_MaxDepth - i);

            for (j = 0; j < m_InferenceRows[i].NumberOfCells; j++)
            {
               m_InferenceRows[i].RowOfCells[j].GrammarNode = NULL;
               m_InferenceRows[i].RowOfCells[j].ReverseComplement = FALSE;
               m_InferenceRows[i].RowOfCells[j].RuleSet = FALSE;
            }
         }
      }
   }
}

/*********************************************************************************/

void GI_CleanUp(void)
{
   ULONG i;
   
   if (m_InferenceRows != NULL)
   {
      for (i = 0; i < m_MaxDepth; i++)
      {
         if (m_InferenceRows[i].RowOfCells != NULL)
         {
            free(m_InferenceRows[i].RowOfCells);
            m_InferenceRows[i].RowOfCells = NULL;
         }
      }
      
      free(m_InferenceRows);
      m_InferenceRows = NULL;
   }
}

/*********************************************************************************/

void GI_GetRootOfTree(CNFADT_HEAD_TYPE *Head, ULONG *HeadDepth)
{
   *HeadDepth = m_MaxDepth - 1;
   *Head = m_InferenceRows[m_MaxDepth - 1].RowOfCells[0].GrammarNode->Entry.Rule.Head;
}

/*********************************************************************************/

void GI_PlaceRuleInCell(ULONG Depth, ULONG ColumnBegin, GI_CELL_INFO_TYPE RuleInfo)
{
   if (Depth >= m_MaxDepth)
   {
      printf("Error: trying to add a rule outside of the matrix dimensions.\n");
      exit(1);      
   }
   else
   {
      if (ColumnBegin >= m_InferenceRows[Depth].NumberOfCells)
      {
         printf("Error: trying to add a rule outside of the matrix dimensions.\n");
         exit(1);      
      }
      else
      {
         m_InferenceRows[Depth].RowOfCells[ColumnBegin] = RuleInfo;
         m_InferenceRows[Depth].RowOfCells[ColumnBegin].RuleSet = TRUE;
      }
   }
}

/*********************************************************************************/

void GI_PreloadType1Rules(CNFADT_DICTIONARY_TYPE *Dictionary)
{
   CNFADT_RULE_AND_METRIC_TYPE NewEntry;
   GI_CELL_INFO_TYPE NewRuleInfo;
   UBYTE i;

   NewEntry.Rule.Type = CNFADT_RULE_TYPE1;

   NewEntry.Metric.InvertedDepth = 0;
   
   NewEntry.Metric.DistanceFromStructuralAddition = 0;
   NewEntry.Metric.NonOverlappingHeadRepeatCount = 0;
   NewEntry.Metric.BodyDepthDifference = 0;      
   NewEntry.Metric.BothBodiesAreStructuralPieces = FALSE;
                  
   VECTORADT_InitializeVector(&(NewEntry.Metric.StructuralComponents));
   VECTORADT_SetElement(&(NewEntry.Metric.StructuralComponents), 1, 0);

   VECTORADT_InitializeVector(&(NewEntry.Metric.ComplementComponents));
   VECTORADT_InitializeVector(&(NewEntry.Metric.SelfComplementComponents));

   VECTORADT_InitializeVector(&(NewEntry.Metric.SumOfComponents));
   VECTORADT_SetElement(&(NewEntry.Metric.SumOfComponents), 1, 0);

   VECTORADT_InitializeVector(&(NewEntry.Metric.AbsDifferenceOfComponents));
   VECTORADT_SetElement(&(NewEntry.Metric.AbsDifferenceOfComponents), 1, 0);

   switch (PARMS_GetSequenceType())
   {
      case PARMS_SEQUENCE_DNA:
      case PARMS_SEQUENCE_RNA:
         for (i = 0; i < INF_NUMBER_OF_PRELOADED_CHARACTERS_DNA_RNA; i++)
         {
            NewEntry.Rule.Body.Type1Body.Label = m_DNARNAPreloadedCharacters[i];
            CNFADT_AddRuleToDictionary(Dictionary, 
                                       NewEntry, 
                                       &(NewRuleInfo.GrammarNode), 
                                       &(NewRuleInfo.ReverseComplement), 
                                       TRUE);
         }
         break;
         
      case PARMS_SEQUENCE_AMINO_ACID:
         for (i = 0; i < INF_NUMBER_OF_PRELOADED_CHARACTERS_AMINO_ACID; i++)
         {
            NewEntry.Rule.Body.Type1Body.Label = m_AminoAcidPreloadedCharacters[i];
            CNFADT_AddRuleToDictionary(Dictionary, 
                                       NewEntry, 
                                       &(NewRuleInfo.GrammarNode), 
                                       &(NewRuleInfo.ReverseComplement), 
                                       TRUE);
         }
         break;
   }
}

/*********************************************************************************/

void GI_CreateType1Rules(CNFADT_DICTIONARY_TYPE *Dictionary,
                         STRINGADT_STRING_TYPE *Sequence,
                         ULONG FirstIndex,
                         ULONG LastIndex)
{
   ULONG i;
   CNFADT_RULE_AND_METRIC_TYPE NewEntry;
   GI_CELL_INFO_TYPE NewRuleInfo;

   for (i = FirstIndex; i <= LastIndex; i++)
   {
      NewEntry.Rule.Type = CNFADT_RULE_TYPE1;
      NewEntry.Rule.Body.Type1Body.Label = Sequence->Characters[i];

      NewEntry.Metric.InvertedDepth = 0;
      
      NewEntry.Metric.DistanceFromStructuralAddition = 0;
      NewEntry.Metric.NonOverlappingHeadRepeatCount = 0;
      NewEntry.Metric.BodyDepthDifference = 0;      
      NewEntry.Metric.BothBodiesAreStructuralPieces = FALSE;

      VECTORADT_InitializeVector(&(NewEntry.Metric.StructuralComponents));
      VECTORADT_SetElement(&(NewEntry.Metric.StructuralComponents), 1, 0);

      VECTORADT_InitializeVector(&(NewEntry.Metric.ComplementComponents));
      VECTORADT_InitializeVector(&(NewEntry.Metric.SelfComplementComponents));

      VECTORADT_InitializeVector(&(NewEntry.Metric.SumOfComponents));
      VECTORADT_SetElement(&(NewEntry.Metric.SumOfComponents), 1, 0);
   
      VECTORADT_InitializeVector(&(NewEntry.Metric.AbsDifferenceOfComponents));
      VECTORADT_SetElement(&(NewEntry.Metric.AbsDifferenceOfComponents), 1, 0);

      CNFADT_AddRuleToDictionary(Dictionary, 
                                 NewEntry, 
                                 &(NewRuleInfo.GrammarNode), 
                                 &(NewRuleInfo.ReverseComplement),
                                 FALSE);

      GI_PlaceRuleInCell(0, i - FirstIndex, NewRuleInfo);
   }
}

/*********************************************************************************/

void GI_CreateType2Rules(CNFADT_DICTIONARY_TYPE *Dictionary,
                         ULONG Depth)
{
   ULONG i;
   ULONG j;
   CNFADT_RULE_AND_METRIC_TYPE MaxEntry;
   CNFADT_RULE_AND_METRIC_TYPE CompareEntry;
   GI_CELL_INFO_TYPE NewRuleInfo;
   BOOLEAN Ignore1;
   CNFADT_NODE_TYPE *Ignore2;
   CNFADT_NODE_TYPE *Ignore3;
   BOOLEAN Found;

   for (i = 0; i < (m_MaxDepth - Depth); i++)
   {
      GI_LoadType2Rule(&MaxEntry, 0, (Depth - 1), i, (i + 1), Depth);

      Found = FALSE;
      if (m_AtLeastOneRepeatAtThisDepth == TRUE)
      {
         Found = CNFADT_FindRuleType2Body(Dictionary, 
                                          MaxEntry.Rule.Body, 
                                          &Ignore1, 
                                          &Ignore2, 
                                          &Ignore3);
      }
      
      for (j = 1; (j < Depth) && (Found == FALSE); j++)
      {
         GI_LoadType2Rule(&CompareEntry, j, (Depth - 1) - j, i, (i + 1) + j, Depth);

         Found = FALSE;
         if (m_AtLeastOneRepeatAtThisDepth == TRUE)
         {
            Found = CNFADT_FindRuleType2Body(Dictionary, 
                                             CompareEntry.Rule.Body, 
                                             &Ignore1, 
                                             &Ignore2, 
                                             &Ignore3);
         }
         
         if ((Found == TRUE) || (GI_SelectRuleAOverRuleB(&CompareEntry, &MaxEntry) == TRUE))
         {
            memcpy (&MaxEntry, &CompareEntry, sizeof(CNFADT_RULE_AND_METRIC_TYPE));
         }
      }
      
      GI_GenerateInitialVectors(&MaxEntry);
      CNFADT_AddRuleToDictionary(Dictionary, 
                                 MaxEntry, 
                                 &(NewRuleInfo.GrammarNode), 
                                 &(NewRuleInfo.ReverseComplement),
                                 FALSE);
      
      GI_PlaceRuleInCell(Depth, i, NewRuleInfo);
   }
}

/*********************************************************************************/

void GI_LoadType2Rule(CNFADT_RULE_AND_METRIC_TYPE *NewEntry,
                        ULONG Row1,
                        ULONG Row2,
                        ULONG Cell1,
                        ULONG Cell2,
                        ULONG Depth)
{
   NewEntry->Rule.Type = CNFADT_RULE_TYPE2;

   NewEntry->Rule.Body.Type2Body[0].Entry = &(m_InferenceRows[Row1].RowOfCells[Cell1].GrammarNode->Entry);
   if (NewEntry->Rule.Body.Type2Body[0].Entry->Metric.SelfReverseComplement == TRUE)
   {
      NewEntry->Rule.Body.Type2Body[0].ReverseComplement = FALSE;
   }
   else
   {
      NewEntry->Rule.Body.Type2Body[0].ReverseComplement = m_InferenceRows[Row1].RowOfCells[Cell1].ReverseComplement;
   }

   NewEntry->Rule.Body.Type2Body[1].Entry = &(m_InferenceRows[Row2].RowOfCells[Cell2].GrammarNode->Entry);
   if (NewEntry->Rule.Body.Type2Body[1].Entry->Metric.SelfReverseComplement == TRUE)
   {
      NewEntry->Rule.Body.Type2Body[1].ReverseComplement = FALSE;
   }
   else
   {
      NewEntry->Rule.Body.Type2Body[1].ReverseComplement = m_InferenceRows[Row2].RowOfCells[Cell2].ReverseComplement;
   }

   NewEntry->Metric.InvertedDepth = Depth;

   NewEntry->Metric.NonOverlappingHeadRepeatCount = 0;

   if (Row1 > Row2)
   {
      NewEntry->Metric.BodyDepthDifference = Row1 - Row2;
   }
   else
   {
      NewEntry->Metric.BodyDepthDifference = Row2 - Row1;
   }

   NewEntry->Metric.ComponentVectorsSet = FALSE;

   NewEntry->Metric.BothBodiesAreStructuralPieces = FALSE;
   NewEntry->Metric.DistanceFromStructuralAddition = 
      NewEntry->Rule.Body.Type2Body[0].Entry->Metric.DistanceFromStructuralAddition +
      NewEntry->Rule.Body.Type2Body[1].Entry->Metric.DistanceFromStructuralAddition;
   if (NewEntry->Metric.DistanceFromStructuralAddition == 0)
   {
      NewEntry->Metric.BothBodiesAreStructuralPieces = TRUE;
   }
   NewEntry->Metric.DistanceFromStructuralAddition += 1;
}

/*********************************************************************************/

void GI_GenerateInitialVectors(CNFADT_RULE_AND_METRIC_TYPE *NewEntry)
{
   if (NewEntry->Metric.ComponentVectorsSet == FALSE)
   {
      NewEntry->Metric.ComponentVectorsSet = TRUE;

      VECTORADT_InitializeVector(&(NewEntry->Metric.StructuralComponents));
      VECTORADT_InitializeVector(&(NewEntry->Metric.ComplementComponents));
      VECTORADT_InitializeVector(&(NewEntry->Metric.SelfComplementComponents));
      VECTORADT_InitializeVector(&(NewEntry->Metric.SumOfComponents));
      VECTORADT_InitializeVector(&(NewEntry->Metric.AbsDifferenceOfComponents));

      if (NewEntry->Rule.Body.Type2Body[0].ReverseComplement == TRUE)
      {
         if (NewEntry->Rule.Body.Type2Body[1].ReverseComplement == TRUE)
         {
            VECTORADT_VectorSum(&(NewEntry->Rule.Body.Type2Body[0].Entry->Metric.ComplementComponents), 
                                 &(NewEntry->Rule.Body.Type2Body[1].Entry->Metric.ComplementComponents), 
                                 &(NewEntry->Metric.StructuralComponents));

            VECTORADT_VectorSum(&(NewEntry->Rule.Body.Type2Body[0].Entry->Metric.StructuralComponents), 
                                 &(NewEntry->Rule.Body.Type2Body[1].Entry->Metric.StructuralComponents), 
                                 &(NewEntry->Metric.ComplementComponents));
         }
         else
         {
            VECTORADT_VectorSum(&(NewEntry->Rule.Body.Type2Body[0].Entry->Metric.ComplementComponents), 
                                 &(NewEntry->Rule.Body.Type2Body[1].Entry->Metric.StructuralComponents), 
                                 &(NewEntry->Metric.StructuralComponents));

            VECTORADT_VectorSum(&(NewEntry->Rule.Body.Type2Body[0].Entry->Metric.StructuralComponents), 
                                 &(NewEntry->Rule.Body.Type2Body[1].Entry->Metric.ComplementComponents), 
                                 &(NewEntry->Metric.ComplementComponents));
         }
      }
      else
      {
         if (NewEntry->Rule.Body.Type2Body[1].ReverseComplement == TRUE)
         {
            VECTORADT_VectorSum(&(NewEntry->Rule.Body.Type2Body[0].Entry->Metric.StructuralComponents), 
                                 &(NewEntry->Rule.Body.Type2Body[1].Entry->Metric.ComplementComponents), 
                                 &(NewEntry->Metric.StructuralComponents));

            VECTORADT_VectorSum(&(NewEntry->Rule.Body.Type2Body[0].Entry->Metric.ComplementComponents), 
                                 &(NewEntry->Rule.Body.Type2Body[1].Entry->Metric.StructuralComponents), 
                                 &(NewEntry->Metric.ComplementComponents));
         }
         else
         {
            VECTORADT_VectorSum(&(NewEntry->Rule.Body.Type2Body[0].Entry->Metric.StructuralComponents), 
                                 &(NewEntry->Rule.Body.Type2Body[1].Entry->Metric.StructuralComponents), 
                                 &(NewEntry->Metric.StructuralComponents));

            VECTORADT_VectorSum(&(NewEntry->Rule.Body.Type2Body[0].Entry->Metric.ComplementComponents), 
                                 &(NewEntry->Rule.Body.Type2Body[1].Entry->Metric.ComplementComponents), 
                                 &(NewEntry->Metric.ComplementComponents));
         }
      }      

      VECTORADT_VectorSum(&(NewEntry->Rule.Body.Type2Body[0].Entry->Metric.SelfComplementComponents), 
                           &(NewEntry->Rule.Body.Type2Body[1].Entry->Metric.SelfComplementComponents), 
                           &(NewEntry->Metric.SelfComplementComponents));

      VECTORADT_VectorSum(&(NewEntry->Metric.StructuralComponents), 
                           &(NewEntry->Metric.ComplementComponents), 
                           &(NewEntry->Metric.SumOfComponents));
      VECTORADT_VectorSum(&(NewEntry->Metric.SumOfComponents), 
                           &(NewEntry->Metric.SelfComplementComponents), 
                           &(NewEntry->Metric.SumOfComponents));

      VECTORADT_VectorAbsDifference(&(NewEntry->Metric.StructuralComponents), 
                                    &(NewEntry->Metric.ComplementComponents), 
                                    &(NewEntry->Metric.AbsDifferenceOfComponents));
      VECTORADT_VectorAbsDifference(&(NewEntry->Metric.AbsDifferenceOfComponents), 
                                    &(NewEntry->Metric.SelfComplementComponents), 
                                    &(NewEntry->Metric.AbsDifferenceOfComponents));
   }
}

/*********************************************************************************/

BOOLEAN GI_SelectRuleAOverRuleB(CNFADT_RULE_AND_METRIC_TYPE *RuleAInfo,
                                 CNFADT_RULE_AND_METRIC_TYPE *RuleBInfo)
{
   BOOLEAN PickRuleA = TRUE;
   BOOLEAN MoveToNextCondition = TRUE;

   /*
      First, pick a rule if d_i + d_j = 0.
   */
   if ((RuleAInfo->Metric.BothBodiesAreStructuralPieces == TRUE) && 
       (RuleBInfo->Metric.BothBodiesAreStructuralPieces == FALSE))
   {
      MoveToNextCondition = FALSE;
   }
   else if ((RuleBInfo->Metric.BothBodiesAreStructuralPieces == TRUE) &&
            (RuleAInfo->Metric.BothBodiesAreStructuralPieces == FALSE))
   {
      MoveToNextCondition = FALSE;
      PickRuleA = FALSE;
   }
   else if ((RuleBInfo->Metric.BothBodiesAreStructuralPieces == TRUE) &&
            (RuleAInfo->Metric.BothBodiesAreStructuralPieces == TRUE))
   {
      if (((RuleBInfo->Rule.Body.Type2Body[0].Entry->Metric.NonOverlappingHeadRepeatCount * 
            (RuleBInfo->Rule.Body.Type2Body[0].Entry->Metric.InvertedDepth + 1)) + 
           (RuleBInfo->Rule.Body.Type2Body[1].Entry->Metric.NonOverlappingHeadRepeatCount * 
            (RuleBInfo->Rule.Body.Type2Body[1].Entry->Metric.InvertedDepth + 1))) >
          ((RuleAInfo->Rule.Body.Type2Body[0].Entry->Metric.NonOverlappingHeadRepeatCount * 
            (RuleAInfo->Rule.Body.Type2Body[0].Entry->Metric.InvertedDepth + 1)) + 
           (RuleAInfo->Rule.Body.Type2Body[1].Entry->Metric.NonOverlappingHeadRepeatCount * 
            (RuleAInfo->Rule.Body.Type2Body[1].Entry->Metric.InvertedDepth + 1))))
      {
         MoveToNextCondition = FALSE;
         PickRuleA = FALSE;
      }
      else if (((RuleBInfo->Rule.Body.Type2Body[0].Entry->Metric.NonOverlappingHeadRepeatCount * 
                 (RuleBInfo->Rule.Body.Type2Body[0].Entry->Metric.InvertedDepth + 1)) + 
                (RuleBInfo->Rule.Body.Type2Body[1].Entry->Metric.NonOverlappingHeadRepeatCount * 
                 (RuleBInfo->Rule.Body.Type2Body[1].Entry->Metric.InvertedDepth + 1))) <
               ((RuleAInfo->Rule.Body.Type2Body[0].Entry->Metric.NonOverlappingHeadRepeatCount * 
                 (RuleAInfo->Rule.Body.Type2Body[0].Entry->Metric.InvertedDepth + 1)) + 
                (RuleAInfo->Rule.Body.Type2Body[1].Entry->Metric.NonOverlappingHeadRepeatCount * 
                 (RuleAInfo->Rule.Body.Type2Body[1].Entry->Metric.InvertedDepth + 1))))
      {
         MoveToNextCondition = FALSE;
      }
   }

   /*
      Second, pick the highest v.
   */
   if (MoveToNextCondition == TRUE)
   {
      GI_GenerateInitialVectors(RuleAInfo);
      GI_GenerateInitialVectors(RuleBInfo);

      switch (VECTORADT_VectorCompare(&(RuleAInfo->Metric.SumOfComponents), 
                                      &(RuleBInfo->Metric.SumOfComponents)))
      {
         case VECTORADT_AGTB:
            MoveToNextCondition = FALSE;
            break;
         
         case VECTORADT_AEQB:
            break;
         
         case VECTORADT_ALTB:
            MoveToNextCondition = FALSE;
            PickRuleA = FALSE;
            break;
      }
   }

   /*
      Third, pick the lowest ||v - v_r| - v_s|.
   */
   if (MoveToNextCondition == TRUE)
   {
      switch (VECTORADT_VectorCompare(&(RuleAInfo->Metric.AbsDifferenceOfComponents), 
                                      &(RuleBInfo->Metric.AbsDifferenceOfComponents)))
      {
         case VECTORADT_AGTB:
            MoveToNextCondition = FALSE;
            PickRuleA = FALSE;
            break;
         
         case VECTORADT_AEQB:
            break;
         
         case VECTORADT_ALTB:
            MoveToNextCondition = FALSE;
            break;
      }
   }

   /*
      Fourth, pick the v associated with the lowest (d_i + d_j).
   */
   if (MoveToNextCondition == TRUE)
   {
      if (RuleBInfo->Metric.DistanceFromStructuralAddition <
            RuleAInfo->Metric.DistanceFromStructuralAddition)
      {
         MoveToNextCondition = FALSE;
         PickRuleA = FALSE;
      }
      else if (RuleBInfo->Metric.DistanceFromStructuralAddition >
               RuleAInfo->Metric.DistanceFromStructuralAddition)
      {
         MoveToNextCondition = FALSE;
      }
   }

   /*
      Fifth, pick the v with the minimum variance tree structure.
   */
   if (MoveToNextCondition == TRUE)
   {
      if (RuleBInfo->Metric.BodyDepthDifference <
            RuleAInfo->Metric.BodyDepthDifference)
      {
         MoveToNextCondition = FALSE;
         PickRuleA = FALSE;
      }
      else if (RuleBInfo->Metric.BodyDepthDifference >
               RuleAInfo->Metric.BodyDepthDifference)
      {
         MoveToNextCondition = FALSE;
      }
   }

   /*
      Sixth, flip a coin.
   */
   if (MoveToNextCondition == TRUE)
   {
      if (COIN_FLIP)
      {
         PickRuleA = FALSE;
      }
   }
   
   return (PickRuleA);
}

/*********************************************************************************/

void GI_AdjustVectorsViaHeadCountAtThisDepth(ULONG Depth, 
                                             CNFADT_DICTIONARY_TYPE *Dictionary)
{
   ULONG i;

   GI_DetermineNonOverlappingHeadRepeats(Depth, Dictionary);

   if (m_AtLeastOneRepeatAtThisDepth == TRUE)
   {
      for (i = 0; i < m_InferenceRows[Depth].NumberOfCells; i++)
      {
         if (m_InferenceRows[Depth].RowOfCells[i].GrammarNode->Entry.Metric.NonOverlappingHeadRepeatCount > 1)
         {
            if (m_InferenceRows[Depth].RowOfCells[i].GrammarNode->Entry.Metric.SelfReverseComplement == TRUE)
            {
               VECTORADT_SetElement(&(m_InferenceRows[Depth].RowOfCells[i].GrammarNode->Entry.Metric.SelfComplementComponents), 1, Depth);
            }
            else
            {
               VECTORADT_SetElement(&(m_InferenceRows[Depth].RowOfCells[i].GrammarNode->Entry.Metric.StructuralComponents), 1, Depth);
            }
            VECTORADT_SetElement(&(m_InferenceRows[Depth].RowOfCells[i].GrammarNode->Entry.Metric.SumOfComponents), 1, Depth);
            VECTORADT_SetElement(&(m_InferenceRows[Depth].RowOfCells[i].GrammarNode->Entry.Metric.AbsDifferenceOfComponents), 1, Depth);
            m_InferenceRows[Depth].RowOfCells[i].GrammarNode->Entry.Metric.DistanceFromStructuralAddition = 0;
         }
      }
   }
}

/*********************************************************************************/

void GI_DetermineNonOverlappingHeadRepeats(ULONG Depth, 
                                          CNFADT_DICTIONARY_TYPE *Dictionary)
{
   ULONG i;
   ULONG k;

   if (m_AtLeastOneRepeatAtThisDepth == TRUE)
   {
      m_AtLeastOneRepeatAtThisDepth = FALSE;
      if (m_InferenceRows[Depth].NumberOfCells > Depth)
      {
         for (i = 0; i < m_InferenceRows[Depth].NumberOfCells; i++)
         {
            if (m_InferenceRows[Depth].RowOfCells[i].GrammarNode->Entry.Metric.HeadRepeatCount > 1)
            {
               if (m_InferenceRows[Depth].RowOfCells[i].GrammarNode->Entry.Metric.NonOverlappingHeadRepeatCount == 0)
               {
                  m_InferenceRows[Depth].RowOfCells[i].GrammarNode->Entry.Metric.NonOverlappingHeadRepeatCount = 1;
                  k = (i + Depth + 1);
                  while (k < m_InferenceRows[Depth].NumberOfCells)
                  {
                     if (m_InferenceRows[Depth].RowOfCells[i].GrammarNode->Entry.Rule.Head ==
                         m_InferenceRows[Depth].RowOfCells[k].GrammarNode->Entry.Rule.Head)
                     {
                        m_InferenceRows[Depth].RowOfCells[i].GrammarNode->Entry.Metric.NonOverlappingHeadRepeatCount += 1;
                        k += (Depth + 1);
                        m_AtLeastOneRepeatAtThisDepth = TRUE;
                     }
                     else
                     {
                        k++;
                     }
                  }
               }
            }
            else
            {
               m_InferenceRows[Depth].RowOfCells[i].GrammarNode->Entry.Metric.NonOverlappingHeadRepeatCount =
                  m_InferenceRows[Depth].RowOfCells[i].GrammarNode->Entry.Metric.HeadRepeatCount;
            }
         }
      }
   }
}

/*********************************************************************************/

void GI_PruneUselessRules(CNFADT_DICTIONARY_TYPE *Dictionary, ULONG Depth, ULONG Cell)
{
   ULONG Row;
   ULONG Col;
   ULONG ColBound;
   ULONG LeftBodyDepth;
   ULONG LeftBodyCell;
   ULONG RightBodyDepth;
   ULONG RightBodyCell;
   
   if (Depth > 1) 
   {
      if (m_InferenceRows[Depth].RowOfCells[Cell].ReverseComplement == TRUE)
      {   
         LeftBodyDepth = m_InferenceRows[Depth].RowOfCells[Cell].GrammarNode->Entry.Rule.Body.Type2Body[1].Entry->Metric.InvertedDepth;
         LeftBodyCell = Cell;
         RightBodyDepth = m_InferenceRows[Depth].RowOfCells[Cell].GrammarNode->Entry.Rule.Body.Type2Body[0].Entry->Metric.InvertedDepth;
         RightBodyCell = (Cell + (Depth - m_InferenceRows[Depth].RowOfCells[Cell].GrammarNode->Entry.Rule.Body.Type2Body[0].Entry->Metric.InvertedDepth));
      }
      else
      {
         LeftBodyDepth = m_InferenceRows[Depth].RowOfCells[Cell].GrammarNode->Entry.Rule.Body.Type2Body[0].Entry->Metric.InvertedDepth;
         LeftBodyCell = Cell;
         RightBodyDepth = m_InferenceRows[Depth].RowOfCells[Cell].GrammarNode->Entry.Rule.Body.Type2Body[1].Entry->Metric.InvertedDepth;
         RightBodyCell = (Cell + (Depth - m_InferenceRows[Depth].RowOfCells[Cell].GrammarNode->Entry.Rule.Body.Type2Body[1].Entry->Metric.InvertedDepth));
      }
   
      for (Row = Depth - 1; Row > 0; Row--)
      {
         if (Row <= LeftBodyDepth)
         {
            Col = LeftBodyCell + (LeftBodyDepth - Row) + 1;
         }
         else
         {
            Col = LeftBodyCell;
         }

         if (Row > RightBodyDepth)
         {
            ColBound = RightBodyCell - (Row - RightBodyDepth) + 1;
         }
         else
         {
            ColBound = RightBodyCell;
         }
         
         while (Col < ColBound)
         {
            CNFADT_DecrementRuleCountInDictionary(Dictionary, &(m_InferenceRows[Row].RowOfCells[Col].GrammarNode));
            m_InferenceRows[Row].RowOfCells[Col].RuleSet = FALSE;
         
            Col++;
         }

      }

      GI_PruneUselessRules(Dictionary, LeftBodyDepth, LeftBodyCell);
      GI_PruneUselessRules(Dictionary, RightBodyDepth, RightBodyCell);
   }
}
