/*****************************************************************/
/*  ELEC 999                                                     */
/*  University of Nebraska-Lincoln                               */
/*  Department of Electrical Engineering                         */
/*  David J. Russell                                             */
/*  5-26-08                                                      */
/*                                                               */
/* 01/30/09 - Added % distribution to output type, version 1.1   */
/* 01/31/09 - Added windowing across an input, version 1.2       */
/* 02/03/09 - Added unpaired repeats and modified paired counts, */
/*             version 1.3                                       */
/* 02/07/09 - Added extreme-value % removal for the stats output,*/
/*             version 1.4                                       */
/* 02/14/09 - Added html output type where text is colored to    */
/*             identify the structural pieces. Modified          */
/*             VectorADT to allow longer vectors, version 1.5    */
/* 02/16/09 - Added minimum ID parameter for labeling nodes on   */
/*             .eps outputs, version 1.6                         */
/* 02/17/09 - Added a parameter to control how many nodes below  */
/*             the root of a structural node are shown on the    */
/*             .eps outputs, version 1.7                         */
/*****************************************************************/

#define _PARMS

/*********************************************************************************/
/* Included Header Files */
/*********************************************************************************/
#include "Globals.h"
#include "Parameters.h"

/*********************************************************************************/
/* Private Module Constants */
/*********************************************************************************/
#define PARMS_STARTING_ARGUMENT_POS                         2
#define PARMS_DELIMITER_POS                                 0
#define PARMS_DELIMITER_CHAR                                '-'
#define PARMS_OPTION_POS                                    1

#define PARMS_MAX_ARGLENGTH                                 80
#define PARMS_MAX_STRING_LENGTH                             80

#define PARMS_DEFAULT_INFILE                                "infile"
#define PARMS_DEFAULT_OUTFILE                               "outfile"
#define PARMS_POST_SCRIPT_EXTENSION                         ".eps"
#define PARMS_GRAMMAR_EXTENSION                             "_grm.txt"
#define PARMS_DISTRIBUTION_EXTENSION                        "_dst.txt"
#define PARMS_HTML_EXTENSION                                ".html"
#define PARMS_DEFAULT_OUTPUT_TYPE                           PARMS_OUTPUT_STRUCTURAL_DISTRIBUTION_PERCENTAGE

#define PARMS_DEFAULT_QUIET_MODE                            FALSE
#define PARMS_DEFAULT_DISABLE_REVERSE_COMPLEMENTS           FALSE

#define PARMS_DEFAULT_PAIRED_DISTANCE_THRESHOLD             35
#define PARMS_DEFAULT_UNPAIRED_DISTANCE_THRESHOLD           100
#define PARMS_DEFAULT_WINDOW_LENGTH                         100
#define PARMS_DEFAULT_DISABLE_WINDOW                        FALSE
#define PARMS_DEFAULT_MINIMUM_STRUCTURAL_LENGTH             3
#define PARMS_DEFAULT_EXTREME_VALUE_REMOVAL_PERCENTAGE      5
#define PARMS_MAX_EXTREME_VALUE_REMOVAL_PERCENTAGE          45

#define PARMS_DEFAULT_MINIMUM_ID_FOR_LABEL                  0

#define PARMS_DEFAULT_ENABLE_SELECTIVE_NODE_OUTPUT          FALSE
#define PARMS_DEFAULT_SELECTIVE_NODE_OUTPUT_PATH_DISTANCE   0

/*********************************************************************************/
/* Private Type Definitions */
/*********************************************************************************/

/*********************************************************************************/
/* Private Module Variables */
/*********************************************************************************/
static BYTE m_InputFileName[PARMS_MAX_STRING_LENGTH];
static BYTE m_OutputFileName[PARMS_MAX_STRING_LENGTH];
static BYTE m_PostScriptOutputFileName[PARMS_MAX_STRING_LENGTH];
static BYTE m_GrammarOutputFileName[PARMS_MAX_STRING_LENGTH];
static BYTE m_DistributionOutputFileName[PARMS_MAX_STRING_LENGTH];
static BYTE m_HTMLOutputFileName[PARMS_MAX_STRING_LENGTH];
static PARMS_OUTPUT_TYPE m_OutputType;
static PARMS_SEQUENCE_TYPE m_SequenceType;
static BOOLEAN m_QuietMode;
static BOOLEAN m_DisableReverseComplements;
static ULONG m_PairedDistanceThreshold;
static ULONG m_UnpairedDistanceThreshold;
static ULONG m_WindowLength;
static BOOLEAN m_DisableWindow;
static ULONG m_MinimumStructuralLength;
static UBYTE m_ExtremeValueRemovalPercentage;
static ULONG m_MinimumIDForLabel;
static BOOLEAN m_EnableSelectiveNodeOutput;
static ULONG m_SelectiveNodeOutputPathDistance;

/*********************************************************************************/
/* Private Module Function Prototypes */
/*********************************************************************************/
static BYTE PARMS_GetNextOption(INT argc, 
                                 BYTE **argv, 
                                 BYTE *OptionDelimiters, 
                                 BYTE *OptionArgument, 
                                 INT *CurrentOptionIndex);
static void PARMS_InitializeDefaultValues(void);
static void PARMS_OutputHeaderText(void);
static void PARMS_OutputHelpText(void);
                           
/*********************************************************************************/
/* Public and Private Module Functions */
/*********************************************************************************/

BYTE *PARMS_GetInputFileName(void)
{
   return (m_InputFileName);
}

/*********************************************************************************/

BYTE *PARMS_GetPostScriptOutputFileName(void)
{
   sprintf(m_PostScriptOutputFileName, "%s%s", m_OutputFileName, PARMS_POST_SCRIPT_EXTENSION);

   return (m_PostScriptOutputFileName);
}

/*********************************************************************************/

BYTE *PARMS_GetGrammarOutputFileName(void)
{
   sprintf(m_GrammarOutputFileName, "%s%s", m_OutputFileName, PARMS_GRAMMAR_EXTENSION);

   return (m_GrammarOutputFileName);
}

/*********************************************************************************/

BYTE *PARMS_GetDistributionOutputFileName(void)
{
   sprintf(m_DistributionOutputFileName, "%s%s", m_OutputFileName, PARMS_DISTRIBUTION_EXTENSION);

   return (m_DistributionOutputFileName);
}

/*********************************************************************************/

BYTE *PARMS_GetHTMLOutputFileName(void)
{
   sprintf(m_HTMLOutputFileName, "%s%s", m_OutputFileName, PARMS_HTML_EXTENSION);

   return (m_HTMLOutputFileName);
}

/*********************************************************************************/

PARMS_OUTPUT_TYPE PARMS_GetOutputType(void)
{
   return (m_OutputType);
}

/*********************************************************************************/

PARMS_SEQUENCE_TYPE PARMS_GetSequenceType(void)
{
   return (m_SequenceType);
}

/*********************************************************************************/

void PARMS_SetSequenceType(PARMS_SEQUENCE_TYPE SequenceType)
{
   switch (SequenceType)
   {
      case PARMS_SEQUENCE_DNA:
      case PARMS_SEQUENCE_RNA:
      case PARMS_SEQUENCE_AMINO_ACID:
         m_SequenceType = SequenceType;
         break;
         
      default:
         printf("Unknown sequence type.\n");
         exit(1);
   }
}

/*********************************************************************************/

BOOLEAN PARMS_GetDisableReverseComplements(void)
{
   return (m_DisableReverseComplements);
}

/*********************************************************************************/

ULONG PARMS_GetPairedDistanceThreshold(void)
{
   return (m_PairedDistanceThreshold);
}

/*********************************************************************************/

ULONG PARMS_GetUnpairedDistanceThreshold(void)
{
   return (m_UnpairedDistanceThreshold);
}

/*********************************************************************************/

ULONG PARMS_GetWindowLength(void)
{
   return (m_WindowLength);
}

/*********************************************************************************/

BOOLEAN PARMS_GetDisableWindow(void)
{
   return (m_DisableWindow);
}

/*********************************************************************************/

ULONG PARMS_GetMinimumStructuralLength(void)
{
   return (m_MinimumStructuralLength);
}

/*********************************************************************************/

UBYTE PARMS_GetExtremeValueRemovalPercentage(void)
{
   return (m_ExtremeValueRemovalPercentage);
}

/*********************************************************************************/

ULONG PARMS_GetMinimumIDForLabel(void)
{
   return (m_MinimumIDForLabel);
}

/*********************************************************************************/

BOOLEAN PARMS_GetEnableSelectiveNodeOutput(void)
{
   return (m_EnableSelectiveNodeOutput);
}

/*********************************************************************************/

ULONG PARMS_GetSelectiveNodeOutputPathDistance(void)
{
   return (m_SelectiveNodeOutputPathDistance);
}

/*********************************************************************************/

BOOLEAN PARMS_ProcessCommandLine(INT argc, BYTE **argv)
{
	BYTE Option;
   BYTE OptionArgument[PARMS_MAX_ARGLENGTH];
   INT CurrentOption = PARMS_STARTING_ARGUMENT_POS;
   BOOLEAN OkToRun = TRUE;
	
   PARMS_InitializeDefaultValues();
   
   while ((Option = PARMS_GetNextOption(argc, argv, "d:e:f:i:L:o:S:T:t:W", OptionArgument, &CurrentOption)) != EOF)
	{
		switch(Option)
		{
			case 'd':
            if (strlen(OptionArgument) > 0)
            {
               m_MinimumIDForLabel = atoi(OptionArgument);
            }
            else
            {
               printf("Unspecified option value.  Use -h for a list of options.\n\n");
               OkToRun = FALSE;
            }
            break;
               
			case 'e':
            if (strlen(OptionArgument) > 0)
            {
               m_ExtremeValueRemovalPercentage = atoi(OptionArgument);
               if (m_ExtremeValueRemovalPercentage > PARMS_MAX_EXTREME_VALUE_REMOVAL_PERCENTAGE)
               {
                  printf("Specified value is greater than the max.  Use -h for a list of options.\n\n");
                  OkToRun = FALSE;
               }
            }
            else
            {
               printf("Unspecified option value.  Use -h for a list of options.\n\n");
               OkToRun = FALSE;
            }
            break;
               
			case 'f':
            if (strlen(OptionArgument) > 0)
            {
               switch(atoi(OptionArgument))
               {
                  case PARMS_OUTPUT_VERBOSE_GRAMMAR_ONLY:
                  case PARMS_OUTPUT_COMPRESSED_GRAMMAR_ONLY:
                  case PARMS_OUTPUT_VERBOSE_GRAMMAR_AND_TREE:
                  case PARMS_OUTPUT_COMPRESSED_GRAMMAR_AND_TREE:
                  case PARMS_OUTPUT_STRUCTURAL_DISTRIBUTION_PERCENTAGE:
                  case PARMS_OUTPUT_HTML_COLORED_SEQUENCE:
                     m_OutputType = atoi(OptionArgument);
                     break;
                     
                  default:
                     printf("Unknown output type.  Use -h for a list of options.\n\n");
                     OkToRun = FALSE;
                     break;
               }
            }
            else
            {
               printf("Unspecified option value.  Use -h for a list of options.\n\n");
               OkToRun = FALSE;
            }
            break;
               
			case 'h':
            PARMS_OutputHelpText();
            OkToRun = FALSE;
				break;

			case 'i':
            if (strlen(OptionArgument) > 0)
            {
               sprintf(m_InputFileName, "%s", OptionArgument);
				}
            else
            {
               printf("No input file specified.  Use -h for a list of options.\n\n");
               OkToRun = FALSE;
            }
            break;

			case 'L':
            if (strlen(OptionArgument) > 0)
            {
               m_MinimumStructuralLength = atoi(OptionArgument);
            }
            else
            {
               printf("Unspecified option value.  Use -h for a list of options.\n\n");
               OkToRun = FALSE;
            }
            break;
               
			case 'o':
            if (strlen(OptionArgument) > 0)
            {
               sprintf(m_OutputFileName, "%s", OptionArgument);
				}
            else
            {
               printf("No output file specified.  Use -h for a list of options.\n\n");
               OkToRun = FALSE;
            }
				break;

         case 'q':
            m_QuietMode = TRUE;
            break;

         case 'r':
            m_DisableReverseComplements = TRUE;
            break;

			case 'S':
            if (strlen(OptionArgument) > 0)
            {
               m_SelectiveNodeOutputPathDistance = atoi(OptionArgument);
            }
            else
            {
               printf("Unspecified option value.  Use -h for a list of options.\n\n");
               OkToRun = FALSE;
            }
            break;
               
         case 's':
            m_EnableSelectiveNodeOutput = TRUE;
            break;

			case 'T':
            if (strlen(OptionArgument) > 0)
            {
               m_PairedDistanceThreshold = atoi(OptionArgument);
            }
            else
            {
               printf("Unspecified option value.  Use -h for a list of options.\n\n");
               OkToRun = FALSE;
            }
            break;

			case 't':
            if (strlen(OptionArgument) > 0)
            {
               m_UnpairedDistanceThreshold = atoi(OptionArgument);
            }
            else
            {
               printf("Unspecified option value.  Use -h for a list of options.\n\n");
               OkToRun = FALSE;
            }
            break;

			case 'W':
            if (strlen(OptionArgument) > 0)
            {
               m_WindowLength = atoi(OptionArgument);
            }
            else
            {
               printf("Unspecified option value.  Use -h for a list of options.\n\n");
               OkToRun = FALSE;
            }
            break;
               
         case 'w':
            m_DisableWindow = TRUE;
            break;

         default:
            printf("Unknown option.  Use -h for a list of options.\n\n");
            OkToRun = FALSE;
            break;
		}
	}

   if (OkToRun == TRUE)
   {
      PARMS_OutputHeaderText();
   }

   return (OkToRun);
}

/*********************************************************************************/

void PARMS_Printf(BYTE *String, BOOLEAN Show)
{
   if (m_QuietMode == FALSE)
   {
      if (Show == TRUE)
      {
         fprintf(stderr, String);
      }
   }
}

/*********************************************************************************/

void PARMS_PrintfProgress(ULONG Current, ULONG Total, BOOLEAN Show)
{
   if (m_QuietMode == FALSE)
   {
      if (Show == TRUE)
      {
         fprintf(stderr, "Progress: %ld / %ld\r", Current, Total);
         if (Current >= Total)
         {
            fprintf(stderr, "\n");
         }
      }
   }
}

/*********************************************************************************/

BYTE PARMS_GetNextOption(INT argc, 
                           BYTE **argv, 
                           BYTE *OptionDelimiters, 
                           BYTE *OptionArgument, 
                           INT *CurrentOptionIndex)
{
	BYTE ReturnOption = EOF;
   BOOLEAN Done;

   OptionArgument[0] = 0;

	if (argc >= PARMS_STARTING_ARGUMENT_POS)
	{
      if (*CurrentOptionIndex <= argc)
      {
   		if (*CurrentOptionIndex < PARMS_STARTING_ARGUMENT_POS)
   		{
   			*CurrentOptionIndex = PARMS_STARTING_ARGUMENT_POS;
   		}

         Done = FALSE;
         while (Done == FALSE)
         {
            if (*CurrentOptionIndex <= argc)
            {
               if (argv[*CurrentOptionIndex - 1][PARMS_DELIMITER_POS] != PARMS_DELIMITER_CHAR)
               {
                  *CurrentOptionIndex += 1;
               }
               else
               {
                  Done = TRUE;
               }
            }
            else
            {
               Done = TRUE;
            }
			}

			if (*CurrentOptionIndex <= argc)
			{
     			ReturnOption = argv[*CurrentOptionIndex - 1][PARMS_OPTION_POS];
     			*CurrentOptionIndex += 1;

     			if (*CurrentOptionIndex <= argc)
     			{
					if (argv[*CurrentOptionIndex - 1][PARMS_DELIMITER_POS] != PARMS_DELIMITER_CHAR)
					{
	     				if (strchr(OptionDelimiters, ReturnOption) != NULL)
	     				{
	     					strcpy(OptionArgument, argv[*CurrentOptionIndex - 1]);
	                  *CurrentOptionIndex += 1;
	     				}
					}
            }
			}
   	}
   }

   return (ReturnOption);
}

/*********************************************************************************/

void PARMS_InitializeDefaultValues(void)
{
   sprintf(m_InputFileName, "%s", PARMS_DEFAULT_INFILE);
   sprintf(m_OutputFileName, "%s", PARMS_DEFAULT_OUTFILE);
   m_SequenceType = PARMS_SEQUENCE_DNA;
   m_DisableReverseComplements = PARMS_DEFAULT_DISABLE_REVERSE_COMPLEMENTS;
   m_OutputType = PARMS_DEFAULT_OUTPUT_TYPE;
   m_PairedDistanceThreshold = PARMS_DEFAULT_PAIRED_DISTANCE_THRESHOLD;
   m_UnpairedDistanceThreshold = PARMS_DEFAULT_UNPAIRED_DISTANCE_THRESHOLD;
   m_WindowLength = PARMS_DEFAULT_WINDOW_LENGTH;
   m_DisableWindow = PARMS_DEFAULT_DISABLE_WINDOW;
   m_MinimumStructuralLength = PARMS_DEFAULT_MINIMUM_STRUCTURAL_LENGTH;
   m_ExtremeValueRemovalPercentage = PARMS_DEFAULT_EXTREME_VALUE_REMOVAL_PERCENTAGE;
   m_MinimumIDForLabel = PARMS_DEFAULT_MINIMUM_ID_FOR_LABEL;
   m_EnableSelectiveNodeOutput = PARMS_DEFAULT_ENABLE_SELECTIVE_NODE_OUTPUT;
   m_SelectiveNodeOutputPathDistance = PARMS_DEFAULT_SELECTIVE_NODE_OUTPUT_PATH_DISTANCE;
}

/*********************************************************************************/

void PARMS_OutputHeaderText(void)
{
   if (m_QuietMode == FALSE)
   {
      printf("\n   ICYK -- ");
      printf("Version %d.%d -- ", VERSION_MAJOR, VERSION_MINOR);
      printf("%s\n", VERSION_DATE);
      printf("   University of Nebraska-Lincoln\n");
      printf("   Department of Electrical Engineering\n");
      printf("   David Russell\n");
      printf("\n");
   }
}

/*********************************************************************************/

void PARMS_OutputHelpText(void)
{
   printf("\n\tUsage: ICYK [options]\n\n");
   printf("\tOptions:\n\n");
   printf("\t-d value\tSpecify minimum inverted depth to receive a label on\n");
   printf("\t\t\t\t.eps tree (default is %d)\n", PARMS_DEFAULT_MINIMUM_ID_FOR_LABEL);
   printf("\t\t\t\t(Ignored for outputs other than %d and %d)\n", PARMS_OUTPUT_VERBOSE_GRAMMAR_AND_TREE, PARMS_OUTPUT_COMPRESSED_GRAMMAR_AND_TREE);
   printf("\t-e value\tSpecify percent of extreme values to remove from average\n");
   printf("\t\t\t\t(default is %d)\n", PARMS_DEFAULT_EXTREME_VALUE_REMOVAL_PERCENTAGE);
   printf("\t\t\t\t(Ignored for outputs other than %d)\n", PARMS_OUTPUT_STRUCTURAL_DISTRIBUTION_PERCENTAGE);
   printf("\t-f value\tSpecify output type (default is %d)\n", PARMS_DEFAULT_OUTPUT_TYPE);
   printf("\t\t\tValues:\t%d = Verbose grammar file only\n", PARMS_OUTPUT_VERBOSE_GRAMMAR_ONLY);
   printf("\t\t\t\t%d = Compressed grammar file only\n", PARMS_OUTPUT_COMPRESSED_GRAMMAR_ONLY);
   printf("\t\t\t\t%d = Verbose grammar file and EPS parse tree\n", PARMS_OUTPUT_VERBOSE_GRAMMAR_AND_TREE);
   printf("\t\t\t\t%d = Compressed grammar file and EPS parse tree\n", PARMS_OUTPUT_COMPRESSED_GRAMMAR_AND_TREE);
   printf("\t\t\t\t%d = Structural distribution percentage\n", PARMS_OUTPUT_STRUCTURAL_DISTRIBUTION_PERCENTAGE);
   printf("\t\t\t\t%d = HTML colored sequence\n", PARMS_OUTPUT_HTML_COLORED_SEQUENCE);
   printf("\t-i filename\tSpecify input FASTA file (default is '%s')\n", PARMS_DEFAULT_INFILE);
   printf("\t-L value\tSpecify minimum structural piece length\n");
   printf("\t\t\t\t(default is %d)\n", PARMS_DEFAULT_MINIMUM_STRUCTURAL_LENGTH);
   printf("\t\t\t\t(Ignored for outputs other than %d and %d)\n", PARMS_OUTPUT_STRUCTURAL_DISTRIBUTION_PERCENTAGE, PARMS_OUTPUT_HTML_COLORED_SEQUENCE);
   printf("\t-o filename\tSpecify output file (default is '%s')\n", PARMS_DEFAULT_OUTFILE);
   printf("\t-q \t\tQuiet mode (default is verbose)\n");
   printf("\t-r \t\tDisable reverse complements in grammar\n");
   printf("\t\t\t\t(default is enabled)\n");
   printf("\t-S value\tSpecify node output path distance (default is %d)\n", PARMS_DEFAULT_SELECTIVE_NODE_OUTPUT_PATH_DISTANCE);
   printf("\t\t\t\t(Ignored for outputs other than %d and %d)\n", PARMS_OUTPUT_VERBOSE_GRAMMAR_AND_TREE, PARMS_OUTPUT_COMPRESSED_GRAMMAR_AND_TREE);
   printf("\t-s \t\tEnable using selective node output\n");
   printf("\t\t\t\t(default is disabled)\n");
   printf("\t\t\t\t(Ignored for outputs other than %d and %d)\n", PARMS_OUTPUT_VERBOSE_GRAMMAR_AND_TREE, PARMS_OUTPUT_COMPRESSED_GRAMMAR_AND_TREE);
   printf("\t-T value\tSpecify paired distance threshold (default is %d)\n", PARMS_DEFAULT_PAIRED_DISTANCE_THRESHOLD);
   printf("\t\t\t\t(Ignored for outputs other than %d and %d)\n", PARMS_OUTPUT_STRUCTURAL_DISTRIBUTION_PERCENTAGE, PARMS_OUTPUT_HTML_COLORED_SEQUENCE);
   printf("\t-t value\tSpecify unpaired distance threshold (default is %d)\n", PARMS_DEFAULT_UNPAIRED_DISTANCE_THRESHOLD);
   printf("\t\t\t\t(Ignored for outputs other than %d and %d)\n", PARMS_OUTPUT_STRUCTURAL_DISTRIBUTION_PERCENTAGE, PARMS_OUTPUT_HTML_COLORED_SEQUENCE);
   printf("\t-W value\tSpecify sequence window length (default is %d)\n", PARMS_DEFAULT_WINDOW_LENGTH);
   printf("\t-w \t\tDisable using a window on the seqeunce\n");
   printf("\t\t\t\t(default is enabled)\n");

   printf("\n");
}
