/*****************************************************************/
/*  ELEC 999                                                     */
/*  University of Nebraska-Lincoln                               */
/*  Department of Electrical Engineering                         */
/*  David J. Russell                                             */
/*  9-21-08                                                      */
/*                                                               */
/* 02/14/09 - Added html output type where text is colored to    */
/*             identify the structural pieces. Modified          */
/*             VectorADT to allow longer vectors, version 1.5    */
/*****************************************************************/

#define _VECTORADT

/*********************************************************************************/
/* Included Header Files */
/*********************************************************************************/
#include "Globals.h"
#include "VectorADT.h"

/*********************************************************************************/
/* Private Module Constants */
/*********************************************************************************/
#define VECTORADT_CHUNK_SIZE                 3

#define VECTORADT_3_ELEMENT_MASK             0x3FEFFBFF

#define VECTORADT_3_ELEMENT_0_MASK           0x000003FF
#define VECTORADT_3_ELEMENT_0_SHIFT          0
#define VECTORADT_3_ELEMENT_0_SIGN           0x00000400
#define VECTORADT_3_ELEMENT_0_ONE            0x00000001
#define VECTORADT_3_ELEMENT_1_MASK           0x000FF800
#define VECTORADT_3_ELEMENT_1_SHIFT          11
#define VECTORADT_3_ELEMENT_1_SIGN           0x00100000
#define VECTORADT_3_ELEMENT_1_ONE            0x00000800
#define VECTORADT_3_ELEMENT_2_MASK           0x3FE00000
#define VECTORADT_3_ELEMENT_2_SHIFT          21
#define VECTORADT_3_ELEMENT_2_SIGN           0x40000000
#define VECTORADT_3_ELEMENT_2_ONE            0x00200000

#define VECTORADT_CELL_INDEX(a)              (a / VECTORADT_CHUNK_SIZE)
#define VECTORADT_CELL_MEMBER(a)             (a % VECTORADT_CHUNK_SIZE)

/*********************************************************************************/
/* Private Type Definitions */
/*********************************************************************************/

/*********************************************************************************/
/* Private Module Function Prototypes */
/*********************************************************************************/
static ULONG VECTORADT_GetElement(VECTORADT_VECTOR_TYPE *Vector, UBYTE CellIndex);

/*********************************************************************************/
/* Private Module Variables */
/*********************************************************************************/

/*********************************************************************************/
/* Public and Private Module Functions */
/*********************************************************************************/

void VECTORADT_InitializeVector(VECTORADT_VECTOR_TYPE *Vector)
{
   Vector->Length = 0;
   memset (Vector->CellsByThree, 0, (sizeof(ULONG) * VECTORADT_MAX_VECTOR_TIMES_THREE));
}

/*********************************************************************************/

void VECTORADT_SetElement(VECTORADT_VECTOR_TYPE *Vector, ULONG Element, UBYTE CellIndex)
{
   UBYTE Index;
   UBYTE Member;

   if (CellIndex < (VECTORADT_MAX_VECTOR_TIMES_THREE * VECTORADT_CHUNK_SIZE))
   {
      Index = VECTORADT_CELL_INDEX(CellIndex);
      Member = VECTORADT_CELL_MEMBER(CellIndex);

      Vector->Length = MAX(Vector->Length,CellIndex);

      switch (Member)
      {
         case 0:
            Vector->CellsByThree[Index] = (Vector->CellsByThree[Index] & ~VECTORADT_3_ELEMENT_0_MASK) | ((Element << VECTORADT_3_ELEMENT_0_SHIFT) & VECTORADT_3_ELEMENT_0_MASK);
            break;

         case 1:
            Vector->CellsByThree[Index] = (Vector->CellsByThree[Index] & ~VECTORADT_3_ELEMENT_1_MASK) | ((Element << VECTORADT_3_ELEMENT_1_SHIFT) & VECTORADT_3_ELEMENT_1_MASK);
            break;

         case 2:
            Vector->CellsByThree[Index] = (Vector->CellsByThree[Index] & ~VECTORADT_3_ELEMENT_2_MASK) | ((Element << VECTORADT_3_ELEMENT_2_SHIFT) & VECTORADT_3_ELEMENT_2_MASK);
            break;
         
         default:
            printf("VectorADT Error: impossible to get here.\n");
            exit(1);
            break;
      }
   }
   else
   {
      printf("VectorADT Error: cell index beyond length of vector.\n");
      exit(1);
   }
}

/*********************************************************************************/

ULONG VECTORADT_GetElement(VECTORADT_VECTOR_TYPE *Vector, UBYTE CellIndex)
{
   UBYTE Index;
   UBYTE Member;
   ULONG Element = 0;

   if (CellIndex < (VECTORADT_MAX_VECTOR_TIMES_THREE * VECTORADT_CHUNK_SIZE))
   {
      Index = VECTORADT_CELL_INDEX(CellIndex);
      Member = VECTORADT_CELL_MEMBER(CellIndex);

      switch (Member)
      {
         case 0:
            Element = ((Vector->CellsByThree[Index] & VECTORADT_3_ELEMENT_0_MASK) >> VECTORADT_3_ELEMENT_0_SHIFT);
            break;

         case 1:
            Element = ((Vector->CellsByThree[Index] & VECTORADT_3_ELEMENT_1_MASK) >> VECTORADT_3_ELEMENT_1_SHIFT);
            break;

         case 2:
            Element = ((Vector->CellsByThree[Index] & VECTORADT_3_ELEMENT_2_MASK) >> VECTORADT_3_ELEMENT_2_SHIFT);
            break;
         
         default:
            printf("Error: cell index beyond length of vector.\n");
            exit(1);
            break;
      }
   }
   else
   {
      printf("VectorADT Error: cell index beyond length of vector.\n");
      exit(1);
   }
   
   return (Element);
}

/*********************************************************************************/

void VECTORADT_VectorSum(VECTORADT_VECTOR_TYPE *A, 
                           VECTORADT_VECTOR_TYPE *B, 
                           VECTORADT_VECTOR_TYPE *C)
{
   UBYTE i;
   UBYTE IndexA = VECTORADT_CELL_INDEX(A->Length);
   UBYTE IndexB = VECTORADT_CELL_INDEX(B->Length);
   
   for (i = 0; i <= MAX(IndexA,IndexB); i++)
   {
      C->CellsByThree[i] = A->CellsByThree[i] + B->CellsByThree[i];

      if ((C->CellsByThree[i] & ~VECTORADT_3_ELEMENT_MASK) != 0)
      {
         printf("VectorADT Error: addition overflow.\n");
         exit(1);
      }
   }

   C->Length = MAX(A->Length,B->Length);
}

/*********************************************************************************/

void VECTORADT_VectorAbsDifference(VECTORADT_VECTOR_TYPE *A, 
                                    VECTORADT_VECTOR_TYPE *B, 
                                    VECTORADT_VECTOR_TYPE *C)
{
   UBYTE i;
   UBYTE IndexA = VECTORADT_CELL_INDEX(A->Length);
   UBYTE IndexB = VECTORADT_CELL_INDEX(B->Length);
   
   for (i = 0; i <= MAX(IndexA,IndexB); i++)
   {
      C->CellsByThree[i] = (A->CellsByThree[i] | ~VECTORADT_3_ELEMENT_MASK) - B->CellsByThree[i];

      switch (C->CellsByThree[i] & (VECTORADT_3_ELEMENT_2_SIGN | VECTORADT_3_ELEMENT_1_SIGN | VECTORADT_3_ELEMENT_0_SIGN))
      {
         case 0:
            C->CellsByThree[i] = (C->CellsByThree[i] ^ (VECTORADT_3_ELEMENT_2_MASK | VECTORADT_3_ELEMENT_1_MASK | VECTORADT_3_ELEMENT_0_MASK)) + (VECTORADT_3_ELEMENT_2_ONE | VECTORADT_3_ELEMENT_1_ONE | VECTORADT_3_ELEMENT_0_ONE);
            break;
            
         case VECTORADT_3_ELEMENT_0_SIGN:
            C->CellsByThree[i] = (C->CellsByThree[i] ^ (VECTORADT_3_ELEMENT_2_MASK | VECTORADT_3_ELEMENT_1_MASK)) + (VECTORADT_3_ELEMENT_2_ONE | VECTORADT_3_ELEMENT_1_ONE);
            break;
            
         case VECTORADT_3_ELEMENT_1_SIGN:
            C->CellsByThree[i] = (C->CellsByThree[i] ^ (VECTORADT_3_ELEMENT_2_MASK | VECTORADT_3_ELEMENT_0_MASK)) + (VECTORADT_3_ELEMENT_2_ONE | VECTORADT_3_ELEMENT_0_ONE);
            break;
            
         case (VECTORADT_3_ELEMENT_1_SIGN | VECTORADT_3_ELEMENT_0_SIGN):
            C->CellsByThree[i] = (C->CellsByThree[i] ^ VECTORADT_3_ELEMENT_2_MASK) + VECTORADT_3_ELEMENT_2_ONE;
            break;
            
         case VECTORADT_3_ELEMENT_2_SIGN:
            C->CellsByThree[i] = (C->CellsByThree[i] ^ (VECTORADT_3_ELEMENT_1_MASK | VECTORADT_3_ELEMENT_0_MASK)) + (VECTORADT_3_ELEMENT_1_ONE | VECTORADT_3_ELEMENT_0_ONE);
            break;
            
         case (VECTORADT_3_ELEMENT_2_SIGN | VECTORADT_3_ELEMENT_0_SIGN):
            C->CellsByThree[i] = (C->CellsByThree[i] ^ VECTORADT_3_ELEMENT_1_MASK) + VECTORADT_3_ELEMENT_1_ONE;
            break;
            
         case (VECTORADT_3_ELEMENT_2_SIGN | VECTORADT_3_ELEMENT_1_SIGN):
            C->CellsByThree[i] = (C->CellsByThree[i] ^ VECTORADT_3_ELEMENT_0_MASK) + VECTORADT_3_ELEMENT_0_ONE;
            break;
            
         case (VECTORADT_3_ELEMENT_2_SIGN | VECTORADT_3_ELEMENT_1_SIGN | VECTORADT_3_ELEMENT_0_SIGN):
            break;
            
         default:
            printf("VectorADT Error: weird problem with bits.\n");
            exit(1);
      }

      C->CellsByThree[i] &= ~(VECTORADT_3_ELEMENT_2_SIGN | VECTORADT_3_ELEMENT_1_SIGN | VECTORADT_3_ELEMENT_0_SIGN);
   }

   C->Length = MAX(A->Length,B->Length);
}

/*********************************************************************************/

VECTORADT_EQUALITY_TYPE VECTORADT_VectorCompare(VECTORADT_VECTOR_TYPE *A, 
                                                VECTORADT_VECTOR_TYPE *B)
{
   VECTORADT_EQUALITY_TYPE Equality = VECTORADT_AEQB;
   UBYTE i;
   UBYTE IndexA = VECTORADT_CELL_INDEX(A->Length);
   UBYTE IndexB = VECTORADT_CELL_INDEX(B->Length);
   
   for (i = MAX(IndexA,IndexB) + 1; (i > 0) && (Equality == VECTORADT_AEQB); i--)
   {
      if (A->CellsByThree[i - 1] > B->CellsByThree[i - 1])
      {
         Equality = VECTORADT_AGTB;
      }
      else if (A->CellsByThree[i - 1] < B->CellsByThree[i - 1])
      {
         Equality = VECTORADT_ALTB;
      }
   }
      
   return (Equality);
}

/*********************************************************************************/

void VECTORADT_PrintVector(VECTORADT_VECTOR_TYPE *Vector, FILE *FileStream)
{
   ULONG j;
   
   fprintf(FileStream, "[");
   for (j = 0; j <= Vector->Length; j++)
   {
      fprintf(FileStream, "%lu", VECTORADT_GetElement(Vector, j));
      if (j < Vector->Length)
      {
         fprintf(FileStream, ",");
      }
      else
      {
         fprintf(FileStream, "]");
      }
   }
}
